# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""nark export reporter base output format class.

- Each output class will support either or both write_facts() and write_report().
"""

import sys

__all__ = ("ReportWriter",)


class ReportWriter(object):
    def __init__(
        self,
        output_b=False,
    ):
        """
        Initiate new instance and open an output file like object.

        Note:
            If you need added bells and wristels (like heading etc.) this would
            probably the method to extend.

        Args:
            path: File-like object or string of path to be opened.

            datetime_format (str): String (sent to strftime) specifying how datetime
                values (Fact start and end) are presented in the output.

            output_b: Whether to open the ``path`` for binary output.
        """
        self.output_b = output_b

    # ***

    def output_setup(
        self,
        output_obj,
        row_limit=0,
        datetime_format=None,
        duration_fmt=None,
    ):
        self.output_file = self.open_output_file(output_obj, self.output_b)

        self.row_limit = row_limit or 0

        self.datetime_format = datetime_format
        if self.datetime_format is None:
            self.datetime_format = "%Y-%m-%d %H:%M:%S"

        self.duration_fmt = duration_fmt
        if self.duration_fmt is None:
            self.duration_fmt = "%H:%M"

    def open_output_file(self, output_obj, output_b=False):
        # FIXME/2020-06-02: Revisit output_b=True, may be different in py3,
        # per these hamster-lib comments:
        #
        #   # No matter through what loops we jump, at the end of the day py27
        #   # ``writerow`` will insist on casting our data to binary str()
        #   # instances. This clearly conflicts with any generic open() that provides
        #   # transparent text input/output and would take care of the encoding
        #   # instead.
        #
        #   # [FIXME]
        #   # If it turns out that this is specific to csv handling we may move it
        #   # there and use a simpler default behaviour for our base method.
        self.output_ours = False
        if not output_obj:
            return sys.stdout
        elif not isinstance(output_obj, str):
            return output_obj
        return self.open_file(output_obj, output_b)

    def open_file(self, path, output_b=False, newline=None):
        self.output_ours = True
        if not output_b:
            return open(path, "w", encoding="utf-8", newline=newline)
        return open(path, "wb")

    # ***

    def write_facts(self, facts):
        """
        Write facts to output file and close the file like object.

        Args:
            facts (Iterable): Iterable of ``nark.Fact`` instances to export.

        Returns:
            None: If everything worked as expected.
        """
        n_written = self.write_facts_list(facts)
        self._close()
        return n_written

    def write_facts_list(self, facts):
        """Write facts to output file."""
        n_written = 0
        for idx, fact in enumerate(facts):
            self._write_fact(idx, fact)
            n_written += 1
            if self.row_limit > 0 and n_written >= self.row_limit:
                break
        return n_written

    def _write_fact(self, idx, fact):
        """
        Represent one ``Fact`` in the output file.

        What this means exactly depends on the format and kind of output.

        Args:
            fact (Fact): The Fact to be written.

        Returns:
            None
        """
        raise NotImplementedError

    # ***

    @property
    def requires_table(self):
        return False

    def write_report(self, table, headers, tabulation=None):
        """
        Write report to output file and close the file like object.

        Args:
            facts (Iterable): Iterable of ``nark.Fact`` instances to export.

        Returns:
            None: If everything worked as expected.
        """
        n_written = self.write_report_table(table, headers, tabulation)
        self._close()
        return n_written

    def write_report_table(self, table, headers, tabulation=None):
        """Write report to output file."""
        n_written = 0
        for row in table:
            self._write_result(row, headers, tabulation)
            n_written += 1
            if self.row_limit > 0 and n_written >= self.row_limit:
                break
        return n_written

    def _write_result(self, row, headers, tabulation=None):
        """
        Represent one ``Fact`` or aggregate result in the output file.

        What this means exactly depends on the format and kind of output
        and search paramaters.

        Args:
            fact (Fact): The Fact to be written.

        Returns:
            None
        """
        raise NotImplementedError

    # ***

    def _close(self):
        """Default teardown method."""
        if self.output_ours:
            self.output_file.close()
        # Rather than clear the file handle, leave set, so caller can inspect.
        #  self.output_file = None


# ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest


class TestActivityManager:
    def test_save_new(self, basestore, activity, mocker):
        """Make sure that saving an new activity calls ``_add``."""
        mocker.patch.object(basestore.activities, "_add", return_value=activity)
        try:
            basestore.activities.save(activity)
        except NotImplementedError:
            pass
        assert basestore.activities._add.called

    def test_save_existing(self, basestore, activity, mocker):
        """Make sure that saving an existing activity calls ``_update``."""
        activity.pk = 0
        mocker.patch.object(basestore.activities, "_update", return_value=activity)
        try:
            basestore.activities.save(activity)
        except NotImplementedError:
            pass
        assert basestore.activities._update.called

    def test_get_or_create_existing(self, basestore, activity, mocker):
        mocker.patch.object(
            basestore.activities,
            "get_by_composite",
            return_value=activity,
        )
        mocker.patch.object(basestore.activities, "save", return_value=activity)
        result = basestore.activities.get_or_create(activity)
        assert result.name == activity.name
        assert basestore.activities.save.called is False

    def test_get_or_create_new(self, basestore, activity, mocker):
        mocker.patch.object(
            basestore.activities,
            "get_by_composite",
            side_effect=KeyError(),
        )
        mocker.patch.object(basestore.activities, "save", return_value=activity)
        result = basestore.activities.get_or_create(activity)
        assert result.name == activity.name
        assert basestore.activities.save.called is True

    def test_add_not_implemented(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities._add(activity)

    def test_update_not_implemented(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities._update(activity)

    def test_remove_not_implemented(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities.remove(activity)

    def test_get_by_composite_not_implemented(self, basestore, activity):
        with pytest.raises(NotImplementedError):
            basestore.activities.get_by_composite(activity.name, activity.category)

    def test_get_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.get(12)

    def test_get_all_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.get_all()

    def test_get_all_by_usage_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.get_all_by_usage()

    def test_gather_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.activities.gather(query_terms=None)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime
from gettext import gettext as _

from ..helpers import fact_time
from ..items.fact import Fact
from . import BaseManager


class BaseFactManager(BaseManager):
    """Base class defining the minimal API for a FactManager implementation."""

    def __init__(self, *args, localize=False, **kwargs):
        super(BaseFactManager, self).__init__(*args, **kwargs)
        # All for one, and one for all, set class-wide behavior.
        Fact.localize(localize)

    # ***

    def save(self, fact, **kwargs):
        """
        Save a Fact to our selected backend.

        Unlike the private ``_add`` and ``_update`` methods, ``save``
        requires that the config given ``fact_min_delta`` is enforced.

        Args:
            fact (nark.Fact): Fact to be saved. Needs to be complete otherwise
            this will fail.

        Returns:
            nark.Fact: Saved Fact.

        Raises:
            ValueError: If ``fact.delta`` is smaller than
              ``self.config['time.fact_min_delta']``
        """

        def _save():
            enforce_fact_min_delta()
            return super(BaseFactManager, self).save(
                fact, cls=Fact, named=False, **kwargs
            )

        def enforce_fact_min_delta():
            # BROKEN/DONT_CARE: (lb): The Facts Carousel does not check the
            # min delta, meaning you could violate fact_min_delta and end up
            # raising from herein. Oh, well, I don't delta, so I don't care.
            if not fact.end:
                # The ongoing, active fact.
                return

            fact_min_delta = int(self.config["time.fact_min_delta"])
            if not fact_min_delta:
                # User has not enabled min-delta behavior.
                return

            min_delta = datetime.timedelta(seconds=fact_min_delta)
            if fact.delta() >= min_delta:
                # Fact is at least as long as user's min-delta.
                return

            message = _(
                "The Fact duration is shorter than the mandatory value of "
                "{} seconds specified in your config.".format(fact_min_delta)
            )
            self.store.logger.error(message)
            raise ValueError(message)

        return _save()

    # ***

    def _add(self, fact):
        """
        Add a new ``Fact`` to the backend.

        Args:
            fact (nark.Fact): Fact to be added.

        Returns:
            nark.Fact: Added ``Fact``.

        Raises:
            ValueError: If passed fact has a PK. New facts should not have one.
            ValueError: If time window is already occupied.
        """
        raise NotImplementedError

    # ***

    def _update(self, fact):
        """
        Update and existing fact with new values.

        Args:
            fact (nark.fact): Fact instance holding updated values.

        Returns:
            nark.fact: Updated Fact

        Raises:
            KeyError: if a Fact with the relevant PK could not be found.
            ValueError: If the the passed activity does not have a PK assigned.
            ValueError: If the time window is already occupied.
        """
        raise NotImplementedError

    # ***

    def remove(self, fact, purge=False):
        """
        Remove a given ``Fact`` from the backend.

        Args:
            fact (nark.Fact): ``Fact`` instance to be removed.

        Returns:
            bool: Success status

        Raises:
            ValueError: If fact passed does not have an pk.
            KeyError: If the ``Fact`` specified could not be found in the backend.
        """
        raise NotImplementedError

    # ***

    def get(self, pk, deleted=None):
        """
        Return a Fact by its primary key.

        Args:
            pk (int): Primary key of the ``Fact to be retrieved``.

            deleted (boolean, optional): False to restrict to non-deleted
                Facts; True to find only those marked deleted; None to find
                all.

        Returns:
            nark.Fact: The ``Fact`` corresponding to the primary key.

        Raises:
            KeyError: If primary key not found in the backend.
        """
        raise NotImplementedError

    # ***

    # Note that get_all() is implemented, by the BaseManager, so it can
    # wrap the 'since' and 'until' inputs, and convert them to datatime
    # objects, before calling gather(), which is implemented by the
    # backends.sqlalchemy.managers GatherFactManager.

    # ***

    def get_all_by_usage(self, query_terms=None, **kwargs):
        """
        Similar to get_all(), but include count of Facts that reference each Activity.
        """
        raise NotImplementedError

    # ***

    def gather(self, query_terms):
        """
        Return a list of ``Facts`` matching given criteria.
        """
        raise NotImplementedError

    # ***

    def get_today(self):
        """
        Return all facts for today, while respecting ``day_start``.

        Returns:
            list: List of ``Fact`` instances.

        Note:
            * This only returns completed facts and does not include the
                active Fact, whether or not it exists.
        """
        self.store.logger.debug(_("Returning today's facts"))

        today = self.store.now.date()
        since = datetime.datetime.combine(today, self.config["time.day_start"])
        until = self.day_end_datetime(today)
        return self.get_all(since=since, until=until)

    def day_end_datetime(self, end_date=None):
        if end_date is None:
            end_date = self.store.now.date()
        start_time = self.config["time.day_start"]
        return fact_time.day_end_datetime(end_date, start_time)

    # ***

    # 2020-01-28: This method only called by nark/tests/. dob replaced its
    # stop-fact mechanism with add_fact, to handle all the possible stop
    # scenarios (e.g., specifying an end time or not; appending the endless
    # fact's description, if it's already got one, etc.).
    def stop_current_fact(self, end_hint=None):
        """
        Stop current 'active fact'.

        Args:
            end_hint (datetime.timedelta or datetime.datetime, optional): Hint to be
                considered when setting ``Fact.end``. If no hint is provided
                ``Fact.end`` will be ``datetime.datetime.now()``. If a ``datetime`` is
                provided, this will be used as ``Fact.end`` value. If a ``timedelta``
                is provided it will be added to ``datetime.datetime.now()``.
                If you want the computed ``end`` to be *before* ``now()``
                you can pass negative ``timedelta`` values. Defaults to None.

        Returns:
            nark.Fact: The stored fact.

        Raises:
            TypeError: If ``end_hint`` is not a ``datetime.datetime`` or
                ``datetime.timedelta`` instance or ``None``.
            ValueError: If there is no currently 'active fact' present.
            ValueError: If the final end value (due to the hint) is before
                the fact's start value.
        """
        self.store.logger.debug(_("Stopping 'active fact'."))

        if not (
            (end_hint is None)
            or isinstance(end_hint, datetime.datetime)
            or isinstance(end_hint, datetime.timedelta)
        ):
            raise TypeError(
                _(
                    "The 'end_hint' you passed needs to be either a"
                    "'datetime.datetime' or 'datetime.timedelta' instance."
                )
            )

        if end_hint:
            if isinstance(end_hint, datetime.datetime):
                end = end_hint
            else:
                end = self.store.now + end_hint
        else:
            end = self.store.now

        # Note that get_current_fact raises KeyError if no Active Fact.
        fact = self.get_current_fact()

        if fact.start > end:
            raise ValueError(
                _(
                    "Cannot end the Fact before it started."
                    " Try editing the Fact instead."
                )
            )
        else:
            fact.end = end
        new_fact = self.save(fact)

        self.store.logger.debug(_("Current fact is now history!"))

        return new_fact

    # ***

    def get_current_fact(self):
        """
        Provide a way to retrieve any existing 'ongoing fact'.

        Returns:
            nark.Fact: An instance representing the <active fact>.

        Raises:
            KeyError: If no ongoing fact is present.
        """

        def _get_current_fact():
            self.store.logger.debug(_("Looking for the 'active fact'."))
            # See alternatively:
            #   facts = self.get_all(endless=True)
            facts = self.endless()
            ensure_one_or_fewer_ongoing(facts)
            ensure_one_or_more_ongoing(facts)
            return facts[0]

        def ensure_one_or_fewer_ongoing(facts):
            if len(facts) <= 1:
                return
            msg = "{} IDs: {}".format(
                _("More than 1 ongoing Fact found. Your database is whacked out!!"),
                ", ".join([str(fact.pk) for fact in facts]),
            )
            self.store.logger.debug(msg)
            raise Exception(msg)

        def ensure_one_or_more_ongoing(facts):
            if facts:
                return
            # See also: NO_ACTIVE_FACT_HELP
            message = _("No active Fact found.")
            self.store.logger.debug(message)
            raise KeyError(message)

        return _get_current_fact()

    # ***

    def find_latest_fact(self, restrict=None):
        assert not restrict or restrict in [
            "ended",
            "ongoing",
        ]
        fact = None
        if not restrict or restrict == "ongoing":
            try:
                fact = self.get_current_fact()
            except KeyError:
                fact = None
            except Exception:
                # (lb): Unexpected! This could mean more than one ongoing Fact found!
                raise
        if fact is None and restrict != "ongoing":
            results = self.get_all(
                sort_cols=("start",),
                sort_orders=("desc",),
                limit=1,
                exclude_ongoing=True,
                # Just to be clear, we want just the Facts.
                include_stats=False,
            )
            if len(results) > 0:
                assert len(results) == 1
                # (lb) Silly, since 2018-06-28 I had `fact, = results`.
                # Was I trying to be clever? Indexing seems more readable.
                fact = results[0]
        return fact

    # ***

    def find_oldest_fact(self):
        fact = None
        results = self.get_all(
            sort_cols=("start",),
            sort_orders=("asc",),
            limit=1,
        )
        if len(results) > 0:
            assert len(results) == 1
            fact = results[0]
        return fact

    # ***

    def cancel_current_fact(self, purge=False):
        """
        Delete the current, ongoing, endless, active Fact.
        (Really just mark it deleted.)

        Returns:
            None: If everything worked as expected.

        Raises:
            KeyError: If no ongoing, active fact found.
        """
        self.store.logger.debug(_("Cancelling 'active fact'."))
        # Note that get_current_fact raises KeyError if no Active Fact.
        fact = self.get_current_fact()
        self.remove(fact, purge)
        return fact

    # ***

    def starting_at(self, fact):
        """
        Return the fact starting at the moment in time indicated by fact.start.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def ending_at(self, fact):
        """
        Return the fact ending at the moment in time indicated by fact.end.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

        Returns:
            nark.Fact: The found Fact, or None if none found.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def antecedent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately preceding the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``start`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The antecedent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        raise NotImplementedError

    # ***

    def subsequent(self, fact=None, ref_time=None):
        """
        Return the Fact immediately following the indicated Fact.

        Args:
            fact (nark.Fact):
                The Fact to reference, with its ``end`` set.

            ref_time (datetime.datetime):
                In lieu of fact, pass the datetime to reference.

        Returns:
            nark.Fact: The subsequent Fact, or None if none found.

        Raises:
            ValueError: If neither ``start`` nor ``end`` is set on fact.
        """
        raise NotImplementedError

    # ***

    def strictly_during(self, start, end, result_limit=1000):
        """
        Return the fact(s) strictly contained within a start and end time.

        Args:
            start (datetime.datetime):
                Start datetime of facts to find.

            end (datetime.datetime):
                End datetime of facts to find.

            result_limit (int):
                Maximum number of facts to find, else log warning message.

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        raise NotImplementedError

    # ***

    def surrounding(self, fact_time):
        """
        Return the fact(s) at the given moment in time.
        Note that this excludes a fact that starts or ends at this time.
        (See antecedent and subsequent for finding those facts.)

        Args:
            fact_time (datetime.datetime):
                Time of fact(s) to match.

        Returns:
            list: List of ``nark.Facts`` instances.

        Raises:
            IntegrityError: If more than one Fact found at given time.
        """
        raise NotImplementedError

    # ***

    def endless(self):
        """
        Return any facts without a fact.start or fact.end.

        Args:
            <none>

        Returns:
            list: List of ``nark.Facts`` instances.
        """
        raise NotImplementedError



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""XML writer output format module."""

import lazy_import

from . import ReportWriter

__all__ = ("XMLWriter",)

# Profiling: load Document: ~ 0.004 secs.
minidom = lazy_import.lazy_module("xml.dom.minidom")


class XMLWriter(ReportWriter):
    """Writer for a basic xml export."""

    # (lb): @elbenfreund noted that XMLWriter copied from 'legacy hamster':
    #   Authored by tstriker <https://github.com/tstriker>. Docstrings by elbenfreund.
    #   https://github.com/projecthamster/hamster/blame/66ed9270c6f0070a4548aca9f070517cc13c85ae
    #       /src/hamster/reports.py#L159
    #   (Other than this class, the nark code authors are either:
    #    landonb (2018-2020); or elbenfreund (2015-2017).)

    def __init__(self, *args, **kwargs):
        """Setup the writer including a main xml document."""
        kwargs["output_b"] = True
        super(XMLWriter, self).__init__(*args, **kwargs)

    def start_document(self, element_name):
        # Profiling: load Document: ~ 0.004 secs.
        self.document = minidom.Document()
        self.fact_list = self.document.createElement(element_name)

    def write_facts(self, facts):
        self.start_document("facts")
        return super(XMLWriter, self).write_facts(facts)

    def _write_fact(self, idx, fact):
        """
        Create new fact element and populate attributes.

        Once the child is prepared append it to ``fact_list``.
        """
        elem = self.document.createElement("fact")
        elem.setAttribute("start", fact.start_fmt(self.datetime_format))
        elem.setAttribute("end", fact.end_fmt(self.datetime_format))
        elem.setAttribute("activity", fact.activity_name)
        elem.setAttribute("duration", fact.format_delta(style=self.duration_fmt))
        elem.setAttribute("category", fact.category_name)
        elem.setAttribute("description", fact.description_or_empty)
        self.fact_list.appendChild(elem)

    def write_report(self, table, headers, tabulation=None):
        self.start_document("results")
        return super(XMLWriter, self).write_report(table, headers, tabulation)

    def _write_result(self, row, headers, tabulation=None):
        """
        Create new fact element and populate attributes.

        Once the child is prepared append it to ``fact_list``.
        """
        fact = self.document.createElement("fact")
        for idx, header in enumerate(headers):
            fact.setAttribute(header, row[idx])
        self.fact_list.appendChild(fact)

    def _close(self):
        """
        Append the xml fact list to the main document write file and cleanup.

        ``toxml`` should take care of encoding everything with UTF-8.
        """

        self.document.appendChild(self.fact_list)
        self.output_file.write(self.document.toxml(encoding="utf-8"))
        return super(XMLWriter, self)._close()



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Shared storage object manager utility functions."""

from sqlalchemy import asc, desc

__all__ = (
    "query_apply_limit_offset",
    "query_apply_true_or_not",
    "query_prepare_datetime",
    "query_sort_order_at_index",
)


def query_apply_limit_offset(query, limit=None, offset=None):
    """
    Applies 'limit' and 'offset' to the database fetch query

    On applies 'limit' if specified; and only applies 'offset' if specified.

    Args:
        query (???): Query (e.g., return from self.store.session.query(...))

        kwargs (keyword arguments):
            limit (int|str, optional): Limit to apply to the query.

            offset (int|str, optional): Offset to apply to the query.

    Returns:
        list: The query passed in, possibly updated with limit and/or offset.
    """
    if limit and limit > 0:
        query = query.limit(limit)
    if offset and offset > 0:
        query = query.offset(offset)
    return query


def query_apply_true_or_not(query, column, condition):
    if condition is not None:
        return query.filter(column == condition)
    return query


def query_sort_order_at_index(sort_orders, idx):
    try:
        direction = desc if sort_orders and sort_orders[idx] == "desc" else asc
    except IndexError:
        direction = asc
    return direction


def query_prepare_datetime(datetm):
    # Be explicit with the format used by the SQL engine, otherwise,
    #   e.g., and_(AlchemyFact.start > start) might match where
    #   AlchemyFact.start == start. In the case of SQLite, the stored
    #   date will be translated with the seconds, even if 0, e.g.,
    #   "2018-06-29 16:32:00", but the datetime we use for the compare
    #   gets translated without, e.g., "2018-06-29 16:32". And we
    #   all know that "2018-06-29 16:32:00" > "2018-06-29 16:32".
    # See also: func.datetime(AlchemyFact.start/end).
    cmp_fmt = "%Y-%m-%d %H:%M:%S"
    text = datetm.strftime(cmp_fmt)
    return text



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from collections import namedtuple
from gettext import gettext as _

from sqlalchemy import case, distinct, func, literal_column
from sqlalchemy.sql.expression import or_

from ....managers.fact import BaseFactManager
from ..objects import (
    AlchemyActivity,
    AlchemyCategory,
    AlchemyFact,
    AlchemyTag,
    fact_tags,
)
from . import query_apply_limit_offset, query_apply_true_or_not, query_prepare_datetime
from .manager_base import BaseAlchemyManager

__all__ = ("GatherFactManager",)


class GatherFactManager(BaseAlchemyManager, BaseFactManager):
    """Fact class aggregate query implementation for FactManager."""

    def __init__(self, *args, **kwargs):
        super(GatherFactManager, self).__init__(*args, **kwargs)

    # ***

    FactStatsTuple = namedtuple(
        "FactStatsTuple",
        (
            "fact",
            "duration",
            "group_count",
            "first_start",
            "final_end",
            "activities",
            "actegories",
            "categories",
            "start_date",
        ),
    )

    RESULT_GRP_INDEX = {
        "duration": 0,
        "group_count": 1,
        "first_start": 2,
        "final_end": 3,
        "activities": 4,
        "actegories": 5,
        "categories": 6,
        "start_date": 7,
    }

    # ***

    def gather(
        self,
        query_terms,
        # - The lazy_tags switch is an internal control that's (as of 2020-05-26)
        #   always False, except for one test that uses it.
        # - It's available if you need to fetch Tag IDs with the tag names.
        #   - Normally when you fetch items in SQLAlchemy, you can reference an
        #     attribute on an item that references data in another table, and
        #     SQLAlchemy (like magic) will query the table and fetch that data.
        #     - E.g., if we fetched from just the Facts table but then accessed
        #       fact.tags on an item, SQLAlchemy runs a SELECT to get those Tags.
        #   - But when building reports, all these extra queries can bog down
        #     the processing. E.g., imagine quickly fetching 10,000 Facts with
        #     one SELECT, but then running 10,000 additional SELECTs just to
        #     get all the tags.
        #     - So this method prefers to fetch tag names in the same query
        #       as fetching the Facts (i.e., keeping the query to just one).
        #       - But this comes at the expense of also fetching tag IDs.
        #         - This method just concatenates tag names for each Fact,
        #           and does not do the extra work of retaining the tag IDs.
        # - tl;dr: Prefer eager loading tag names over lazy-loading Tag items,
        #   so that exporting and reporting is fast.
        #   - But if you (eventually) need Tag IDs associated with Facts
        #     fetched by this method, enable lazy_tags (so called because
        #     SQLAlchemy will lazy-load the Tag item when you access a Fact's
        #     fact.tags attribute).
        lazy_tags=False,
    ):
        """
        Return matching facts, maybe each with stats, given some search criteria.

        If no search criteria is given, returns all facts.

        Args:
            query_terms (nark.managers.query_terms.QueryTerms, required):
                The requested query settings used to find Activities,
                and also the requested results settings. See the QueryTerms
                class for details.

            lazy_tags: Set True to skip concatenating tag names for each Fact,
                but to instead have the actual Tag items lazy-loaded upon
                accessing each fact.tags in the results.

        Returns:
            list: A list of matching item instances or (item, *statistics) tuples.
        """
        qt = query_terms

        magic_tag_sep = "%%%%,%%%%"

        add_aggregates = qt.include_stats or qt.is_grouped or qt.sorts_cols_has_stat
        # Cannot request lazy_tags when grouping, just not how it works.
        # (If we did, then, say, if group_tags, we'd need to outerjoin
        # fact_tags and AlchemyFact, or we'd need to add the tags_subquery
        # anyway. But lazy_tags is meant for lazy-loading an individual
        # Fact's tags, which has no meaning on aggregate results, so nip
        # that now.)
        if add_aggregates and lazy_tags:
            errmsg = _("Cannot request lazy_tags when grouping results.")
            raise Exception(errmsg)

        def _get_all_facts():
            self.store.logger.debug(qt)

            must_support_db_engine_funcs()

            query = self.store.session.query(AlchemyFact)

            query, tags_subquery = _get_all_prepare_tags_subquery(query)

            query, span_cols = _get_all_prepare_span_cols(query)

            query, actg_cols = _get_all_prepare_actg_cols(query)

            query, start_date = _get_all_prepare_start_date(query)

            query = _get_all_prepare_joins(query)

            query = self.query_filter_by_fact_times(
                query,
                qt.since,
                qt.until,
                qt.endless,
                qt.partial,
            )

            query = self.query_filter_by_activities(query, qt)

            query = self.query_filter_by_categories(query, qt)

            query = query_filter_by_search_term(query)

            query = self.query_filter_by_item_pk(query, AlchemyFact, qt.key)

            query = query_apply_true_or_not(query, AlchemyFact.deleted, qt.deleted)

            query = _get_all_filter_by_ongoing(query)

            query = query_group_by_aggregate(query, tags_subquery)

            has_facts = True
            query = self.query_order_by_sort_cols(
                query,
                qt,
                has_facts,
                span_cols,
                start_date,
                tags_subquery,
            )

            query = query_apply_limit_offset(query, qt.limit, qt.offset)

            query = query_select_with_entities(
                query,
                span_cols,
                actg_cols,
                start_date,
                tags_subquery,
            )

            self.query_prepared_trace(query)

            if qt.count_results:
                results = query.count()
            else:
                # Profiling: 2018-07-15: (lb): ~ 0.120 s. to fetch latest of 20K Facts.
                records = query.all()
                results = _gather_process_results(records)

            return results

        # ***

        def must_support_db_engine_funcs():
            if self.store.config["db.engine"] == "sqlite":
                return

            errmsg = _(
                "This feature does not work with the current DBMS engine: ‘{}’."
                " (Please tell the maintainers if you want this supported!"
                " That, or switch to SQLite to use this feature.)".format(
                    self.store.config["db.engine"]
                )
            )
            raise NotImplementedError(errmsg)

        # ***

        def _gather_process_results(records):
            if not records:
                return records
            elif not qt.include_stats and not add_aggregates and lazy_tags:
                return _gather_process_facts_only(records)
            return _gather_process_facts_and_aggs(records)

        # The list of results returned to the user is one of:
        # - A list of raw AlchemyFact objects;
        # - A list of hydrated Fact objects (or of a caller-specified subclass); or
        # - A list of tuples comprised of the AlchemyFact or Fact, followed by
        #   a number of calculated, aggregate columns (added when grouping).
        #   - The order of items in each tuple is determined by the function:
        #       query_select_with_entities
        #     which calls query.with_entities with a list that starts with:
        #       AlchemyFact
        #     and then adds aggregate columns from the functions:
        #       _get_all_prepare_span_cols, and
        #       _get_all_prepare_actg_cols.
        #   - The order of aggregates is also reflected by RESULT_GRP_INDEX.
        # - The intermediate results might also end with a coalesced Tags value
        #   (see _get_all_prepare_tags_subquery), but the tags_subquery.tags_col
        #   is pulled before the results are returned.

        def _gather_process_facts_only(records):
            if qt.raw:
                return records

            # Because not add_aggregates, results are single items, AlchemyFact.
            # Note also that we ignore qt.named_tuples here (which does not
            # apply unless also qt.include_stats, which is not True here).
            records = [fact.as_hamster(self.store) for fact in records]
            return records

        def _gather_process_facts_and_aggs(records):
            results = []
            # PROFILING: Here's a loop over all the results!
            # If the user didn't limit or restrict their query,
            # this could be all the Facts!
            for fact, *cols in records:
                fact_or_tuple = _process_record(fact, cols)
                results.append(fact_or_tuple)
            return results

        def _process_record(fact, cols):
            new_tags = _process_record_tags(cols)
            new_fact = _process_record_prepare_fact(fact, new_tags)
            _process_record_reduce_aggregates(cols)
            return _process_record_new_fact_or_tuple(new_fact, cols)

        # +++

        def _process_record_tags(cols):
            # If tags were fetched, they'll be coalesced in the final column.
            new_tags = None
            if cols and not lazy_tags:
                tags = cols.pop()
                new_tags = tags.split(magic_tag_sep) if tags else []
            return new_tags

        # +++

        def _process_record_prepare_fact(fact, new_tags):
            # Unless the caller wants raw results, create a Fact.
            if not qt.raw:
                # Create a new, first-class Fact (or FactDressed). And if
                # the results are aggregate, create a frequency distribution,
                # or number of uses per tag (stored at tag.freq).
                return fact.as_hamster(
                    self.store,
                    new_tags,
                    set_freqs=qt.is_grouped,
                )

            # Even if user wants raw results, still attach the tags.
            if new_tags:
                assert not lazy_tags  # Just FYI.
                # Note that the group_concat reduced the Tag names to a single
                # string that we pulled apart into a list, new_tags, while we
                # need to convert to AlchemyTag to assign to the AlchemyFact.
                # CAVEAT: These tags are without their IDs! This is fine for
                #         reporting, but if you need Tag PKs, set lazy_tags=True.
                pkless_tags = [
                    AlchemyTag(pk=None, name=name, deleted=False, hidden=False)
                    for name in new_tags
                ]
                fact.tags = pkless_tags

            return fact

        # +++

        def _process_record_reduce_aggregates(cols):
            if not cols or not qt.include_stats:
                return

            _process_record_reduce_aggregate_activities(cols)
            _process_record_reduce_aggregate_actegories(cols)
            _process_record_reduce_aggregate_categories(cols)

        def _process_record_reduce_aggregate_activities(cols):
            i_activities = GatherFactManager.RESULT_GRP_INDEX["activities"]
            _process_record_reduce_aggregate_value(cols, i_activities)

        def _process_record_reduce_aggregate_actegories(cols):
            i_actegories = GatherFactManager.RESULT_GRP_INDEX["actegories"]
            _process_record_reduce_aggregate_value(cols, i_actegories)

        def _process_record_reduce_aggregate_categories(cols):
            i_categories = GatherFactManager.RESULT_GRP_INDEX["categories"]
            _process_record_reduce_aggregate_value(cols, i_categories)

        def _process_record_reduce_aggregate_value(cols, index):
            encoded_value = cols[index]
            # Note that group_concat generates None if all values were None.
            if encoded_value == 0:
                return

            if encoded_value:
                concated_values = encoded_value.split(magic_tag_sep)
                unique_values = set(concated_values)
            else:
                unique_values = ""
            cols[index] = unique_values

        # +++

        def _process_record_new_fact_or_tuple(new_fact, cols):
            # Make a tuple with the calculated and group-by aggregates,
            # if any, when requested by the caller.
            if qt.include_stats:
                if not qt.named_tuples:
                    return [new_fact] + cols
                else:
                    return GatherFactManager.FactStatsTuple(new_fact, *cols)
            return new_fact

        # ***

        def _get_all_prepare_tags_subquery(query):
            if lazy_tags and not qt.match_tags:
                return query, None

            tags_subquery = query
            # (lb): Always include tags. We could let SQLAlchemy lazy load,
            # but this can be slow. E.g., on 15K Facts, calling fact.tags on
            # each -- triggering lazy load -- takes 7 seconds on my machine.
            # As opposed to 0 seconds (rounded down) when preloading tags.
            # - Note earlier must_support_db_engine_funcs() b/c SQLite-specific.
            tags_col = func.group_concat(
                AlchemyTag.name,
                magic_tag_sep,
            ).label("facts_tags")
            tags_subquery = tags_subquery.add_columns(tags_col)
            # (lb): Leaving this breadcrumb for now; feel free to delete later.
            # - sqlalchemy 1.2.x allowed ambiguous joins and would use the first
            #   instance of such a join for its ON clause. But sqlalchemy 1.3.x
            #   requires that you be more specific. Here's the original code:
            #       tags_subquery = tags_subquery.outerjoin(fact_tags)
            #   and following is (obviously) the new code. Note that this was
            #   the only place I found code that needed fixing, but I would not
            #   be surprised to find more. Hence this note-to-self, for later.
            tags_subquery = tags_subquery.outerjoin(
                fact_tags,
                AlchemyFact.pk == fact_tags.columns.fact_id,
            )
            tags_subquery = tags_subquery.outerjoin(AlchemyTag)

            if qt.match_tags:
                tags_subquery = self.query_filter_by_tags(tags_subquery, qt)

            tags_subquery = tags_subquery.group_by(AlchemyFact.pk)

            # (lb): 2019-01-22: Old comment re: joinedload. Leaving here as
            # documentation in case I try using joinedload again in future.
            #   # FIXME/2018-06-25: (lb): Not quite sure this'll work...
            #   # http://docs.sqlalchemy.org/en/latest/orm/loading_relationships.html
            #   #   # joined-eager-loading
            #   from sqlalchemy.orm import joinedload
            #   # 2019-01-22: Either did not need, or did not work, !remember which!
            #   tags_subquery = tags_subquery.options(joinedload(AlchemyFact.tags))
            tags_subquery = tags_subquery.with_entities(AlchemyFact.pk, tags_col)
            tags_subquery = tags_subquery.subquery("tag_names")
            query = query.join(tags_subquery, AlchemyFact.pk == tags_subquery.c.pk)

            return query, tags_subquery

        # ***

        def _get_all_prepare_span_cols(query):
            if not add_aggregates:
                return query, None

            span_cols = []

            query, group_span_col = _get_all_prepare_span_cols_group_span(query)
            span_cols.append(group_span_col)

            query, group_count_col = _get_all_prepare_span_cols_group_count(query)
            span_cols.append(group_count_col)

            query, first_start_col = _get_all_prepare_span_cols_first_start(query)
            span_cols.append(first_start_col)

            query, final_end_col = _get_all_prepare_span_cols_final_end(query)
            span_cols.append(final_end_col)

            return query, span_cols

        def _get_all_prepare_span_cols_group_span(query):
            # For most Facts, we could calculate the time window span with
            # simple end-minus-start math, e.g.,
            #   func.julianday(AlchemyFact.end)
            #    - func.julianday(AlchemyFact.start)
            # But this would miss the final ongoing, active Fact. So check
            # first if end is None, and use the 'now' time if so.
            endornow_col = case(
                [(AlchemyFact.end != None, AlchemyFact.end)],  # noqa: E711
                else_=query_prepare_datetime(self.store.now),
            )

            span_col = _get_all_prepare_span_cols_group_span_dbms_specific(endornow_col)

            group_span_col = func.sum(span_col).label("duration")
            query = query.add_columns(group_span_col)
            return query, group_span_col

        def _get_all_prepare_span_cols_group_span_dbms_specific(endornow_col):
            # MAYBE/2020-05-15: Implement this feature for other DBMS engines.
            # - The julianday function is SQLite-specific.
            #   - A "pure" fix might mean doing the calculation after getting
            #     all the results (i.e., post-processing the SQL response).
            #   - (lb): But I like how easy julianday is. It works, and I personally
            #     have no stake in using another DBMS engine. We can add additional
            #     support as needed (i.e., as users' non-SQLite interests dictate).
            if self.store.config["db.engine"] == "sqlite":
                return _get_all_prepare_span_cols_group_span_sqlite(endornow_col)
            else:  # pragma: no cover
                # See exception thrown by must_support_db_engine_funcs() if not SQLite.
                assert False  # Not reachable.

        def _get_all_prepare_span_cols_group_span_sqlite(endornow_col):
            span_col = func.julianday(endornow_col) - func.julianday(AlchemyFact.start)
            return span_col

        def _get_all_prepare_span_cols_group_count(query):
            group_count_col = func.count(distinct(AlchemyFact.pk)).label("group_count")
            query = query.add_columns(group_count_col)
            return query, group_count_col

        def _get_all_prepare_span_cols_first_start(query):
            first_start_col = func.min(AlchemyFact.start).label("first_start")
            query = query.add_columns(first_start_col)
            return query, first_start_col

        def _get_all_prepare_span_cols_final_end(query):
            final_end_col = func.max(AlchemyFact.end).label("final_end")
            query = query.add_columns(final_end_col)
            return query, final_end_col

        # ***

        # Note that grouping by Activity.pk inherently also groups by the
        # Category, because of the many-to-one relationship.
        # - The same is also true if grouping by Activity.pk and Category.pk,
        #   which is essentially no different than grouping by just Activity.pk.
        # - As such, we make the distinction between grouping by just Activity
        #   and by both Activity and Category by using the Activity name when
        #   grouping only by the Activity, but using the Activity ID when the
        #   Category is also involved.
        # - When we group by Activity name, it combines Facts from same-named
        #   Activities in different Categories. As an example, if a user has
        #   Email@Personal and Email@Work, grouping by Activity name will show
        #   overall time spent on Email (which they could also achieve with,
        #   say, an #Email tag).
        # - Also note that grouping by Category (but not Activity) will
        #   collapse one or more Activities, as will grouping by Tags.
        # - Depending on what's grouped -- Activity, Category, and/or Tags --
        #   we'll make a coalesced Activities or Act@gories column to combine
        #   all Activity names that are grouped, or a Categories column to
        #   report all the Categories of each set of grouped Facts (each row).

        def _get_all_prepare_actg_cols(query):
            actg_cols = None
            if not add_aggregates:
                return query, actg_cols

            # - Use placeholder values -- may as well be zero -- for columns we do
            #   not need for this query, so that the return tuple is always the same
            #   size and always has the same layout as reflected by RESULT_GRP_INDEX.
            #   Because the concatenated names columns are strings, the caller can
            #   check if the value is not a string, but an integer (0) instead, to
            #   help decide which columns to use in the report output.
            # - NOTE: Use literal_column() to avoid warning on with_entities:
            #     /home/user/.virtualenvs/dob/lib/python3.8/site-packages/
            #       sqlalchemy/orm/query.py:217:
            #         SADeprecationWarning: Plain string expression passed to Query()
            #         should be explicitly declared using literal_column();
            #         automatic coercion of this value will be removed in SQLAlchemy 1.4
            activities_col = literal_column("0")
            actegories_col = literal_column("0")
            categories_col = literal_column("0")

            if qt.group_activity and qt.group_category:
                # The Activity@Category for each result is unique/not an aggregate.
                pass
            elif qt.group_activity:
                # One Activity per result, but one or more Categories were flattened.
                query, categories_col = _get_all_prepare_actg_cols_categories(query)
            elif qt.group_category:
                # One Category per result, but one or more Activities were flattened.
                query, activities_col = _get_all_prepare_actg_cols_activities(query)
            elif qt.group_tags or qt.group_days:
                # When grouping by tags, both Activities and Categories are grouped.
                query, actegories_col = _get_all_prepare_actg_cols_actegories(query)

            actg_cols = [activities_col, actegories_col, categories_col]

            return query, actg_cols

        def _get_all_prepare_actg_cols_activities(query):
            activities_col = func.group_concat(
                AlchemyActivity.name,
                magic_tag_sep,
            ).label("facts_activities")
            query = query.add_columns(activities_col)
            return query, activities_col

        def _get_all_prepare_actg_cols_actegories(query):
            # SQLite supports column || concatenation, which is + in SQLAlchemy.
            # MAYBE/2020-05-18: Is there a config value that specs the '@' sep?
            # - I.e., replace the hardcoded '@' with a config value.
            actegory_col = AlchemyActivity.name + "@" + AlchemyCategory.name
            # SKIP/Not necessary:
            #   actegory_col.label("actegory")
            #   query = query.add_columns(actegory_col)
            #   actg_cols.append(actegory_col)
            actegories_col = func.group_concat(
                actegory_col,
                magic_tag_sep,
            ).label("facts_actegories")
            query = query.add_columns(actegories_col)
            return query, actegories_col

        def _get_all_prepare_actg_cols_categories(query):
            categories_col = func.group_concat(
                AlchemyCategory.name,
                magic_tag_sep,
            ).label("facts_categories")
            query = query.add_columns(categories_col)
            return query, categories_col

        # ***

        # MAYBE/2020-05-19: Splice Facts at midnight so that each
        # day is reported as 24 hours, rather than reported as the
        # sum of all the complete Facts that start on the day (in
        # which case, a Fact that starts the day before that runs
        # into the day will not be counted (undercount); and a Fact
        # that starts on the day by ends on the next will be have
        # all its time counted (over-count).

        def _get_all_prepare_start_date(query):
            # If we were not going to return the start_date with the results,
            # rather than checking `not add_aggregates`, we would instead
            # check `not qt.group_days` and return if so.
            if not add_aggregates:
                return query, None

            # FIXME/MAYBE/2020-05-19: Prepare other time-grouped reports. See:
            #   https://www.sqlite.org/lang_datefunc.html
            # - MAYBE: Prepare monthly reports using 'start of month':
            #     SELECT date('now', 'start of month');
            #     if group_months: ...
            # - MAYBE: Prepare weekly reports using `weekday N` (0=Sunday);
            #     if group_weeks: ...
            # - MAYBE: Prepare Sprint reports using... not sure.
            #     if group_sprint???

            start_date = func.date(
                AlchemyFact.start,
            ).label("start_date")
            query = query.add_columns(start_date)
            if qt.group_days:
                query = query.group_by(start_date)

            return query, start_date

        # ***

        def _get_all_prepare_joins(query):
            join_category = (
                qt.is_grouped  # b/c _get_all_prepare_actg_cols_categories
                or qt.match_categories
                or qt.search_terms
                or qt.sort_cols_has_any("activity")
                or qt.sort_cols_has_any("category")
            )
            if add_aggregates or qt.match_activities or join_category:
                # Equivalent: AlchemyFact.activity or AlchemyActivity.
                query = query.outerjoin(AlchemyFact.activity)
            if join_category:
                # Equivalent: AlchemyActivity.category or AlchemyCategory.
                query = query.outerjoin(AlchemyActivity.category)
            return query

        # ***

        def query_filter_by_search_term(query):
            """
            Limit query to facts that match the search terms.

            Terms are matched against ``Category.name`` and ``Activity.name``.
            The matching is not case-sensitive.
            """
            if not qt.search_terms:
                return query

            filters = []
            for term in qt.search_terms:
                filters.append(AlchemyFact.description.ilike("%{}%".format(term)))
                if qt.broad_match:
                    filters.append(AlchemyActivity.name.ilike("%{}%".format(term)))
                    filters.append(AlchemyCategory.name.ilike("%{}%".format(term)))
                    filters.append(AlchemyTag.name.ilike("%{}%".format(term)))
            query = query.filter(or_(*filters))

            return query

        # ***

        def _get_all_filter_by_ongoing(query):
            if not qt.exclude_ongoing:
                return query
            if qt.exclude_ongoing:
                query = query.filter(AlchemyFact.end != None)  # noqa: E711
            return query

        # ***

        def query_group_by_aggregate(query, tags_subquery):
            if not qt.is_grouped:
                # Need to group by Fact.pk because some aggregates, e.g.,
                # COUNT(), will want to collapse all rows.
                return query_group_by_pk(query)
            return query_group_by_meta(query, tags_subquery)

        def query_group_by_pk(query):
            # Group by Fact.pk, as there might be "duplicate" Facts because
            # of the fact_tags join. The 'facts_tags' column coalesces Tags.
            query = query.group_by(AlchemyFact.pk)
            return query

        def query_group_by_meta(query, tags_subquery):
            query = query_group_by_activity_and_category(query)
            query = query_group_by_tags(query, tags_subquery)
            return query

        def query_group_by_activity_and_category(query):
            if qt.group_activity and qt.group_category:
                # NOTE:The wrong group-by returns 1 record:
                #        query = query.group_by(AlchemyFact.activity)
                #      generates:
                #         GROUP BY activities.id = facts.activity_id
                #      But we want more simply:
                #         GROUP BY activities.id
                query = query.group_by(AlchemyActivity.pk)
                # Each Activity is associated with exactly one Category,
                # so the group_by(AlchemyActivity.pk) is sufficient, but
                # for completeness, group on the Category, too.
                query = query.group_by(AlchemyCategory.pk)
            elif qt.group_activity:
                # This is kinda smudgy. What does it mean to group by the Activity?
                # - Do you mean an Activity in the traditional sense, which is
                #   Activity@Category? (in which case, group-by the Activity.pk,
                #   and include the Category in the grouping).
                # - Or is this an Activity is a looser sense, as in an Activity
                #   name? (in which case, group by the Activity name, which means
                #   two Activities with the same name but in different Categories
                #   become grouped).
                # To be the most flexible, if the user wants to group by the
                # Activity but does not also specify the Category, then, sure,
                # group by the Activity name. They can add the Category grouping
                # if they want to group on the actual Activity@Category.
                query = query.group_by(AlchemyActivity.name)
            elif qt.group_category:
                query = query.group_by(AlchemyCategory.pk)
            return query

        def query_group_by_tags(query, tags_subquery):
            if not qt.group_tags:
                return query

            assert not lazy_tags
            return query.group_by(tags_subquery.c.facts_tags)

        # ***

        def query_select_with_entities(
            query,
            span_cols,
            actg_cols,
            start_date,
            tags_subquery,
        ):
            # Even if grouping, we still want to fetch all columns. For one,
            # _process_results expects a Fact object as leading item in each
            # result tuple, and also because as_hamster expects certain fields
            # (and in a specific order). We'd also have to at least specify
            # AlchemyFact.pk so that SQLAlchemy uses `FROM facts`, and not,
            # e.g., `FROM category`. So use all Fact cols to start the select.
            columns = [AlchemyFact]

            # The order of the columns added here is reflected by RESULT_GRP_INDEX.
            # - Note also if `add_aggregates` is True, then both span_cols and
            #   actg_cols should be not None. But we'll check for None-ness
            #   just to be safe.
            if span_cols is not None:
                columns.extend(span_cols)
            if actg_cols is not None:
                columns.extend(actg_cols)
            if add_aggregates:
                columns.append(start_date)

            # Ensure tags_subquery.tags_col is last, because
            # _process_record_tags expects (pops) it.
            if tags_subquery is not None:
                assert not lazy_tags
                outer_tags_col = func.group_concat(
                    tags_subquery.c.facts_tags,
                    magic_tag_sep,
                ).label("facts_tags")
                columns.append(outer_tags_col)

            query = query.with_entities(*columns)

            return query

        # ***

        return _get_all_facts()

    # ***

    def query_order_by_sort_col(
        self,
        query,
        query_terms,
        sort_col,
        direction,
        has_facts,
        # The following columns are specific to a Fact gather.
        span_cols,
        start_date,
        tags_subquery,
    ):
        qt = query_terms

        assert has_facts

        if sort_col == "start" or not sort_col:
            query = self.query_order_by_start(query, direction)
        elif sort_col == "time":
            i_duration = GatherFactManager.RESULT_GRP_INDEX["duration"]
            query = query.order_by(direction(span_cols[i_duration]))
        elif sort_col == "day":
            assert start_date is not None
            query = query.order_by(direction(start_date))
        elif sort_col == "activity":
            # If grouping by only category, this sort does not work: The
            # activity names are group_concat'enated into the 'activities'
            # column, which must be post-processed -- split on magic_tag_sep,
            # made unique, and sorted. Such a dob command might look like this:
            #   `dob list facts --group category --sort activity`
            # But if also grouping by activity, or tags, order-by here works.
            # So sorting activity when grouping category is done by caller.
            if (
                qt.group_activity
                or qt.group_tags
                or qt.group_days
                or not qt.group_category
            ):
                query = query.order_by(direction(AlchemyActivity.name))
        elif sort_col == "category":
            # If sorting by category but grouping by activity (or tags), the
            # caller must sort during post-processing, after transforming the
            # categories aggregate (which get_all returns as a set) into an
            # alphabetically ordered string. (We could sort here, but it
            # would have no effect, so might as well not.)
            if (
                qt.group_category
                or qt.group_tags
                or qt.group_days
                or not qt.group_activity
            ):
                query = query.order_by(direction(AlchemyCategory.name))
        elif sort_col == "tag" and tags_subquery is not None:
            # Don't sort by the aggregate column, because tags aren't
            # sorted in the aggregate (they're not even unique, it's
            # just a long string built from all the tags).
            # - So this won't sort the table:
            #     query = query.order_by(direction(tags_subquery))
            # But because we checked tags_subquery is not None, we know
            # that the Tag table is joined -- so we can sort by the tag
            # name. Except if grouping by activity or category, then the
            # sort won't stick, so skip it in that case.
            if not qt.group_activity and not qt.group_category:
                query = query.order_by(direction(tags_subquery.c.facts_tags))
        elif sort_col == "usage" and span_cols is not None:
            i_group_count = GatherFactManager.RESULT_GRP_INDEX["group_count"]
            query = query.order_by(direction(span_cols[i_group_count]))
        elif sort_col == "name":
            # It makes sense to sort Activities, Categories, and Tags by their
            # names, but a Fact does not have a name. So what does `--sort name`
            # mean to a Fact? 2020-05-19: This code had been ignoring --sort name
            # (all sort options are shared, so it's not considered wrong to
            # receive such a request), but we could treat `--sort name` as a
            # request to sort by description. There's not another mechanism to
            # sort by description, and it seems pretty useless, anyway, except
            # maybe to group Facts without a description. So we might as well wire
            # sorting by name to sorting be description, rather than ignoring it.
            query = query.order_by(direction(AlchemyFact.description))
        elif sort_col == "fact":
            # (lb): There is (or at least should be) no meaning with Fact IDs,
            # i.e., you should think of them as UUID values, and not having any
            # relationship to one another, other than as an indicator of identity.
            # So ordering by the PK, just because it happens to be an integer
            # and can be compared against other PKs, is possible, but it also
            # doesn't accomplish much. I suppose you could sort by PK to ensure that
            # you can recreate a report in the same order as before, if that's your
            # thing. 2020-05-18: In any case, the `--sort fact` option has existed
            # in the CLI, but the code here has been ignoring it (and just treating
            # it like `--sort start`). But there are no reasons, other than perhaps
            # ideological, that we cannot at least wire it to the PK. So here ya go:
            query = query.order_by(direction(AlchemyFact.pk))
        else:
            self.store.logger.warning("Unknown sort_col: {}".format(sort_col))
        return query

    # ***


# ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures to help test nark.helpers and related."""

import datetime

# CAVEAT: Note that freeze_time expected to be used with this!
#  @freeze_time('2015-12-25 18:00')
factoid_fixture = (
    ("raw_fact", "time_hint", "expectation"),
    [
        # Use clock-to-clock format, the date inferred from now; with actegory.
        (
            "13:00 to 16:30: foo@bar",
            "verify_both",
            {
                "start_raw": "13:00",
                "end_raw": "16:30",
                "start": datetime.datetime(2015, 12, 25, 13, 0, 0),
                "end": datetime.datetime(2015, 12, 25, 16, 30, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # *** Test the '-' and ' - ' separators.
        # Use clock-to-clock without spaces format.
        (
            "13:00-16:30: foo@bar",
            "verify_both",
            {
                "start_raw": "13:00",
                "end_raw": "16:30",
                "start": datetime.datetime(2015, 12, 25, 13, 0, 0),
                "end": datetime.datetime(2015, 12, 25, 16, 30, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Test wrap-around relative clock times specified.
        (
            "12:00:11 - 11:01:59: act @",
            "verify_both",
            {
                "start_raw": "12:00:11",
                "end_raw": "11:01:59",
                "start": datetime.datetime(2015, 12, 25, 12, 0, 11),
                "end": datetime.datetime(2015, 12, 26, 11, 1, 59),
                "activity": "act",
                "category": "",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # 2020-12-22: Note that "datish-datish" works (no spaces around dash)
        # if there's only a single dash in the datetime portion being parsed.
        # Note freeze_time("2015-12-25 18:00") is Friday, so 2020-12-21 start.
        (
            "Monday-13:00: foo@bar",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end_raw": "13:00",
                "start": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end": datetime.datetime(2015, 12, 21, 13, 0, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        (
            "Monday - 13:00: foo@bar",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end_raw": "13:00",
                "start": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end": datetime.datetime(2015, 12, 21, 13, 0, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # ***
        # Use datetime-to-datetime format, with actegory.
        (
            "2015-12-12 13:00 to 2015-12-12 16:30: foo@bar",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": datetime.datetime(2015, 12, 12, 16, 30, 0),
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": datetime.datetime(2015, 12, 12, 16, 30, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # The end date is inferred from start date.
        (
            "2015-12-12 13:00 - 18:00 foo@bar",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": "18:00",
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": datetime.datetime(2015, 12, 12, 18, 00, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # actegory spanning day (straddles) midnight) and spanning multiple days.
        (
            "2015-12-12 13:00 - 2015-12-25 18:00 foo@bar",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": datetime.datetime(2015, 12, 25, 18, 00, 0),
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": datetime.datetime(2015, 12, 25, 18, 00, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Create open/ongoing/un-ended fact.
        (
            "2015-12-12 13:00 foo@bar",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # - LOST_IT: (lb): I had a note suggesting this test tickles that unlikely path
        #   with start=None (in must_create_fact_from_factoid, or mend_fact_timey_wimey,
        #   or Fact.create_from_factoid), but I cannot find that branch, nor does this
        #   test tickle any affirm(False)... but not deleting this comment, because I'm
        #   still suspicious.
        # Create ongoing fact starting at right now.
        (
            "foo@bar",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Similar to previous, but avoid re-raising ParserMissingDatetimeOneException
        # (and grabs us one more line of coverage, the `if minmax[0] == 0` branch).
        (
            "foo@bar",
            "verify_then",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        (
            "foo@bar",
            "verify_still",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # ***Test verify_both's `strictly_two` branches.
        # Test the first branch:
        #   elif strictly_two:
        #       self.raise_missing_datetime_two()
        (
            "2015-12-12 13:00 foo@bar",
            "verify_both",
            {
                "err": "Expected to find the two datetimes separated by one of: ",
            },
        ),
        # Test the second branch of same, immediately following previous.
        (
            "2015-12-12 13:00: foo@bar",
            "verify_both",
            {
                "err": "Expected to find the two datetimes separated by one of: ",
            },
        ),
        # *** Test tags.
        # 2 Simple Tags.
        (
            "2015-12-12 13:00 foo@bar: #precious #hashish, i like ike",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": ["precious", "hashish"],
                "description": "i like ike",
                "warnings": [],
            },
        ),
        # Multiple Tags are identified by a clean leading delimiter character.
        (
            '2015-12-12 13:00 foo@bar, #just walk away "#not a tag", blah',
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": ['just walk away "#not a tag"'],
                "description": "blah",
                "warnings": [],
            },
        ),
        # Alternative tag delimiter; and quotes are just consumed as part of tag.
        (
            '2015-12-12 13:00 foo@bar, #just walk away @"totes a tag", blah',
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": ["just walk away", '"totes a tag"'],
                "description": "blah",
                "warnings": [],
            },
        ),
        # Test '#' in description, elsewhere, after command, etc.
        (
            '2015-12-12 13:00 baz@bat",'
            " #tag1, #tag2 tags cannot come #too late, aha!"
            " Time is also ignored at end: 12:59",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "baz",
                "category": 'bat"',
                "tags": ["tag1"],
                "description": "#tag2 tags cannot come #too late, aha!"
                " Time is also ignored at end: 12:59",
                "warnings": [],
            },
        ),
        # ***
        # No tags, just description, tests Parser.RE_SPLIT_CAT_AND_TAGS.match
        # returns None.
        (
            "2015-12-12 13:00 foo@bar: blah blah blah",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "blah blah blah",
                "warnings": [],
            },
        ),
        # No act@cat but tags and description, tests `parse_tags_and_remainder`
        # self.re_item_sep.match branch.
        (
            "2015-12-12 13:00: #baz: bat",
            "verify_start",
            {
                "err": "Expected to find an Activity name.",
                # If lenient=True:
                #  "start_raw": datetime.datetime(2015, 12, 12, 13, 0, 0),
                #  "end_raw": None,
                #  "start": datetime.datetime(2015, 12, 12, 13, 0, 0),
                #  "end": None,
                #  "activity": "",
                #  "category": "",
                #  "tags": ["baz"],
                #  "description": "bat",
                #  "warnings": [],
            },
        ),
        # (lb): This one... may or may be making the best decision:
        # what looks like a tag is being gobbled as the description.
        (
            "foo@bar #baz",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "#baz",
                "warnings": [],
            },
        ),
        # (lb): Similar to previous example, this might not be what user
        # would except, but adding a separator (colon) does signal intent,
        # so this one feels like it does behave the way it should.
        (
            "foo@bar #baz:",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar #baz",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        (
            "foo@bar #baz: #bat",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar #baz",
                "tags": ["bat"],
                "description": "",
                "warnings": [],
            },
        ),
        (
            "foo@bar #baz: bat",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar #baz",
                "tags": [],
                "description": "bat",
                "warnings": [],
            },
        ),
        (
            "foo@bar #baz: #bat: cat",
            "verify_none",
            {
                "start_raw": None,
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar #baz",
                "tags": ["bat"],
                "description": "cat",
                "warnings": [],
            },
        ),
        # ***
        # Note that dateparser uses current clock time for time relative to
        # now, e.g., 'yesterday', '1 day ago', etc, but it uses midnight for
        # absolute references, e.g., 'Monday', 'January', etc.
        # - The 18:00:00 is in reference to: @freeze_time('2015-12-25 18:00').
        (
            "Monday until 2 days ago: act @",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end_raw": datetime.datetime(2015, 12, 23, 18, 0, 0),
                "start": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end": datetime.datetime(2015, 12, 23, 18, 0, 0),
                "activity": "act",
                "category": "",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # ***
        # Test must_parse_datetimes_known `not self.raw_datetime1` if-branch.
        (
            "-90: foo@bar",
            "verify_end",
            {
                # For dob, because dob.facts.add_fact.add_fact raises
                "err_inclusive": (
                    "Please specify `start` to add Fact before time existed."
                ),
                #
                # For nark, after nark.helpers.parsing.parse_factoid
                "start_raw": None,
                "end_raw": "-90",
                "start": None,
                # 90 mins. before 6p Xmas 2015.
                "end": datetime.datetime(2015, 12, 25, 16, 30, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Note that you should not include the time verb
        # (which is encapsulated in the time_hint).
        (
            "to -90: foo@bar",
            "verify_end",
            {
                # For dob, because dob.facts.add_fact.add_fact raises
                "err_inclusive": (
                    "Please specify `start` to add Fact before time existed."
                ),
                #
                # For nark, after nark.helpers.parsing.parse_factoid
                "start_raw": None,
                "end_raw": datetime.datetime(1990, 12, 25, 0, 0, 0),
                "start": None,
                "end": datetime.datetime(1990, 12, 25, 0, 0, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Test must_parse_datetimes_known `not self.raw_datetime1` if-branch.
        (
            "-120 to -90: foo@bar",
            "verify_both",
            {
                "start_raw": "-120",
                "end_raw": "-90",
                #  I.e., 120 and 90 minutes before now.
                "start": datetime.datetime(2015, 12, 25, 16, 0, 0),
                "end": datetime.datetime(2015, 12, 25, 16, 30, 0),
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Test must_parse_datetimes_known `not self.raw_datetime1` else-branch.
        (
            "Monday: foo@bar",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 21, 0, 0, 0),
                "end": None,
                "activity": "foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Test warn_if_datetime_missing_clock_time, if looks_like_clock_abbrev.
        (
            "2015-12-12 300 foo@bar",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 0, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 0, 0, 0),
                "end": None,
                "activity": "300 foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [
                    "The identified datetime is missing the time of day. "
                    "Perhaps those three digits after the date "
                    "should be a 4-digit clocktime?"
                ],
                "err": "",
            },
        ),
        # Test warn_if_datetime_missing_clock_time, if not looks_like_clock_abbrev.
        (
            "2015-12-12 9 foo@bar",
            "verify_start",
            {
                "start_raw": datetime.datetime(2015, 12, 12, 0, 0, 0),
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 12, 0, 0, 0),
                "end": None,
                "activity": "9 foo",
                "category": "bar",
                "tags": [],
                "description": "",
                "warnings": [
                    "The identified datetime is missing the time of day. "
                    "Is that what you wanted? (Probably not!)"
                ],
                "err": "",
            },
        ),
        (
            "+10m to @",
            "verify_start",
            {
                # For dob, because dob.facts.add_fact.add_fact raises
                "err_inclusive": "Start after end!",
                #
                # For nark, after nark.helpers.parsing.parse_factoid
                "start_raw": "+10m",
                "end_raw": None,
                "start": datetime.datetime(2015, 12, 25, 18, 10, 0),
                "end": None,
                "activity": "to",
                "category": "",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # Obscure Factoid covers last branch in `must_parse_datetime_from_rest`.
        # Note that start=+ time makes sense in context on import, where there
        # will be a previous Fact to reference, but in this fixture, this Fact
        # stands alone, so supply relative end to avoid default end = now.
        (
            "+10m to +20m foo@",
            "verify_both",
            {
                "start_raw": "+10m",
                "end_raw": "+20m",
                "start": datetime.datetime(2015, 12, 25, 18, 10, 0),
                # Note the +20m is relative to start, not now (which was used on start).
                "end": datetime.datetime(2015, 12, 25, 18, 30, 0),
                "activity": "foo",
                "category": "",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
        # *** must_parse_datetimes_magic tests.
        # Test `if two_is_okay`-`elif strictly_two` in `must_parse_datetimes_magic`.
        (
            "+10m foo@bar",
            "verify_both",
            {
                "err": "Expected to find the two datetimes separated by one of: ",
            },
        ),
        # Test `if not self.raw_datetime1`-`== 'verify_end'`
        # in `must_parse_datetimes_magic`.
        (
            "Monday foo@bar",
            "verify_end",
            {
                "err": "Expected to find the two datetimes separated by one of: ",
            },
        ),
        # Test `if not self.raw_datetime1`-`!= 'verify_end'`
        # in `must_parse_datetimes_magic`.
        (
            "Monday foo@bar",
            "verify_start",
            {
                "err": "Expected to find a datetime.",
            },
        ),
        (
            "Tuesday foo@bar",
            "verify_start",
            {
                # Raises ParserMissingDatetimeOneException.
                "err": "Expected to find a datetime.",
                # Otherwise:
                #  'start_raw': 'Tuesday',
                #  'end_raw': None,
                #  # 'start': datetime.datetime(2015, 12, 24, 18, 0, 0),
                #  'start': datetime.datetime(2015, 12, 25, 18, 0, 0),
                #  'end': None,
                #  'activity': 'foo',
                #  'category': 'bar',
                #  'tags': [],
                #  'description': '',
                #  'warnings': [],
            },
        ),
        # Cover final line in must_parse_datetimes_magic.
        (
            "Tuesday to 12:00: foo@",
            "verify_both",
            {
                "start_raw": datetime.datetime(2015, 12, 22, 0, 0, 0),
                "end_raw": "12:00",
                # freeze_time is 2015-12-25 18:00 Friday; Dec. 22 is Tuesday.
                "start": datetime.datetime(2015, 12, 22, 0, 0, 0),
                "end": datetime.datetime(2015, 12, 22, 12, 0, 0),
                "activity": "foo",
                "category": "",
                "tags": [],
                "description": "",
                "warnings": [],
            },
        ),
    ],
)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from ....managers.category import BaseCategoryManager
from ..objects import AlchemyCategory, AlchemyFact
from . import query_apply_true_or_not
from .manager_base import BaseAlchemyManager

__all__ = ("CategoryManager",)


class CategoryManager(BaseAlchemyManager, BaseCategoryManager):
    """ """

    def __init__(self, *args, **kwargs):
        super(CategoryManager, self).__init__(*args, **kwargs)

    # ***

    def get_or_create(self, category, raw=False, skip_commit=False):
        """
        Custom version of the default method in order to provide access
        to alchemy instances.

        Args:
            category (nark.Category): Category we want.
            raw (bool): Wether to return the AlchemyCategory instead.

        Returns:
            nark.Category or None: Category.
        """

        self.store.logger.debug("Received: {!r} / raw: {}".format(category, raw))

        try:
            category = self.get_by_name(category.name, raw=raw)
        except KeyError:
            category = self._add(category, raw=raw, skip_commit=skip_commit)
        return category

    # ***

    def _add(self, category, raw=False, skip_commit=False):
        """
        Add a new category to the database.

        This method should not be used by any client code. Call ``save`` to
        make the decission wether to modify an existing entry or to add a new
        one is done correctly.

        Args:
            category (nark.Category): nark Category instance.
            raw (bool): Wether to return the AlchemyCategory instead.

        Returns:
            nark.Category: Saved instance, as_hamster()

        Raises:
            ValueError: If the name to be added is already present in the db.
            ValueError: If category passed already got an PK. Indicating that
                update would be more appropriate.
        """
        self.adding_item_must_not_have_pk(category)

        alchemy_category = AlchemyCategory(
            pk=None,
            name=category.name,
            deleted=bool(category.deleted),
            hidden=bool(category.hidden),
        )

        result = self.add_and_commit(
            alchemy_category,
            raw=raw,
            skip_commit=skip_commit,
        )

        return result

    # ***

    def _update(self, category):
        """
        Update a given Category.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            nark.Category: Updated category.

        Raises:
            ValueError: If the new name is already taken.
            ValueError: If category passed does not have a PK.
            KeyError: If no category with passed PK was found.
        """

        self.store.logger.debug("Received: {!r}".format(category))

        if not category.pk:
            message = _(
                "The Category passed ('{!r}') does not have a PK."
                " We do not know which entry to modify."
            ).format(category)
            self.store.logger.error(message)
            raise ValueError(message)
        alchemy_category = self.store.session.query(AlchemyCategory).get(category.pk)
        if not alchemy_category:
            message = _("No Category with PK ‘{}’ was found.").format(category.pk)
            self.store.logger.error(message)
            raise KeyError(message)
        alchemy_category.name = category.name

        try:
            self.store.session.commit()
        except IntegrityError as err:
            message = _(
                "An error occured! Is category.name already present in the database?"
                " / Error: '{}'."
            ).format(str(err))
            self.store.logger.error(message)
            raise ValueError(message)

        return alchemy_category.as_hamster(self.store)

    # ***

    def remove(self, category):
        """
        Delete a given category.

        Args:
            category (nark.Category): Category to be removed.

        Returns:
            None: If everything went alright.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            ValueError: If category passed does not have an pk.
        """

        self.store.logger.debug("Received: {!r}".format(category))

        if not category.pk:
            message = _(
                "Category ‘{!r}’ has no PK. Are you trying to remove a new Category?"
            ).format(category)
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_category = self.store.session.query(AlchemyCategory).get(category.pk)
        if not alchemy_category:
            message = _("No Category with PK ‘{}’ was found.").format(category.pk)
            self.store.logger.error(message)
            raise KeyError(message)

        self.store.session.delete(alchemy_category)
        self.store.session.commit()
        self.store.logger.debug("Deleted: {!r}".format(category))

    # ***

    def get(self, pk, deleted=None):
        """
        Return a category based on their pk.

        Args:
            pk (int): PK of the category to be retrieved.

        Returns:
            nark.Category: Category matching given PK.

        Raises:
            KeyError: If no such PK was found.

        Note:
            We need this for now, as the service just provides pks, not names.
        """

        self.store.logger.debug("Received PK: ‘{}’".format(pk))

        if deleted is None:
            result = self.store.session.query(AlchemyCategory).get(pk)
        else:
            query = self.store.session.query(AlchemyCategory)
            query = query.filter(AlchemyCategory.pk == pk)
            query = query_apply_true_or_not(query, AlchemyCategory.deleted, deleted)
            results = query.all()
            assert len(results) <= 1
            result = results[0] if results else None

        if not result:
            message = _("No Category with PK ‘{}’ was found.").format(pk)
            self.store.logger.error(message)
            raise KeyError(message)
        self.store.logger.debug("Returning: {!r}".format(result))
        return result.as_hamster(self.store)

    # ***

    def get_by_name(self, name, raw=False):
        """
        Return a category based on its name.

        Args:
            name (str): Unique name of the category.
            raw (bool): Whether to return the AlchemyCategory instead.

        Returns:
            nark.Category: Category of given name.

        Raises:
            KeyError: If no category matching the name was found.

        """
        self.store.logger.debug("Received name: ‘{}’ / raw: {}".format(name, raw))

        try:
            result = (
                self.store.session.query(AlchemyCategory).filter_by(name=name).one()
            )
        except NoResultFound:
            message = _("No Category named ‘{}’ was found.").format(name)
            self.store.logger.debug(message)
            raise KeyError(message)

        if not raw:
            result = result.as_hamster(self.store)
            self.store.logger.debug("Returning: {!r}.".format(result))
        return result

    # ***
    # *** gather() call-outs (used by get_all/get_all_by_usage).
    # ***

    @property
    def _gather_query_alchemy_cls(self):
        return AlchemyCategory

    @property
    def _gather_query_order_by_name_col(self):
        return "category"

    def _gather_query_requires_fact(self, qt, compute_usage):
        requires_fact_table = super(CategoryManager, self)._gather_query_requires_fact(
            qt, compute_usage
        )

        # Ensure _gather_query_start_aggregate called -- and that we
        # select from Fact, joined Activity, joined Category -- if
        # the user is matching on or ordering by Activity.
        requires_fact_table = (
            requires_fact_table
            or qt.match_activities
            or qt.sort_cols_has_any("activity")
        )

        return requires_fact_table

    def _gather_query_start_aggregate(self, qt, agg_cols):
        query = self.store.session.query(AlchemyFact, *agg_cols)
        query = query.outerjoin(AlchemyFact.activity)
        query = query.outerjoin(AlchemyCategory)
        return query

    # ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from unittest.mock import patch

import pytest
from freezegun import freeze_time

from nark.helpers.parsing import (
    Parser,
    ParserMissingActivityException,
    ParserMissingDatetimeTwoException,
    ParserMissingSeparatorActivity,
    parse_factoid,
)
from nark.tests.helpers.conftest import factoid_fixture

# See also parser testing via Fact.create_from_factoid in tests/items/test_fact.py
# and also Factoid parsing start/end resolution in dob/tests/fact/test_add_fact.py.


@pytest.fixture
def parser():
    return Parser()


class TestParser(object):
    def test_parser_to_str(self, parser):
        assert str(parser).startswith("raw: ")

    def test_parser_factoid_None(self, parser):
        """Test that Factoid parser fails if Activity not indicated."""
        # That is, a Factoid is at least an Activity.
        with pytest.raises(ParserMissingActivityException):
            parser.dissect_raw_fact(factoid=None)

    def test_parser_factoid_list(self, parser):
        parser.dissect_raw_fact(factoid=["01:00 to 03:00 act @"])

    def test_parser_factoid_activity_with_sep(self, parser):
        parser.dissect_raw_fact(factoid="yesterday until 03:00 act @")

    def test_parser_factoid_activity_without_sep(self, parser):
        with pytest.raises(ParserMissingSeparatorActivity):
            parser.dissect_raw_fact(factoid="yesterday: act", time_hint="verify_end")

    # Test single entry DATE_TO_DATE_SEPARATORS__RAW.
    @patch("nark.helpers.parsing.DATE_TO_DATE_SEPARATORS__RAW", ["to"])
    def test_parser_factoid_missing_two_single_sep(self, parser):
        with pytest.raises(ParserMissingDatetimeTwoException):
            parser.dissect_raw_fact("13:00: foo@bar", "verify_both")

    @freeze_time("2015-12-25 18:00")
    @pytest.mark.parametrize(*factoid_fixture)
    def test_helpers_parsing_parse_factoid(
        self,
        raw_fact,
        time_hint,
        expectation,
    ):
        """
        Make sure that a valid raw fact creates a proper Fact.

        CPYST: Test the factoid_fixture's:

            pytest --pdb -s -vv \
        tests/helpers/test_parsing.py::TestParser::test_helpers_parsing_parse_factoid
        """
        fact_dict, err = parse_factoid(
            raw_fact,
            time_hint=time_hint,
            # Use lenient=True so this returns errors, rather than raise.
            lenient=True,
        )
        if "err" in expectation and expectation["err"]:
            assert str(err).startswith(expectation["err"])
        else:
            assert fact_dict["start"] == expectation["start_raw"]
            assert fact_dict["end"] == expectation["end_raw"]
            assert fact_dict["activity"] == expectation["activity"]
            assert fact_dict["category"] == expectation["category"]
            assert fact_dict["tags"] == expectation["tags"]
            assert fact_dict["description"] == expectation["description"]
            assert fact_dict["warnings"] == expectation["warnings"]



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from icalendar import Calendar


class TestICALWriter(object):
    """Make sure the iCal writer works as expected."""

    def test_ical_writer_init(self, ical_writer):
        """Make sure that init creates a new calendar instance to add events to."""
        assert ical_writer.calendar

    def test_ical_writer_write_facts_expected(self, ical_writer, fact, mocker):
        """Make sure that the fact attached to the calendar matches our expectations."""
        mocker.patch.object(ical_writer.calendar, "add_component")
        ical_writer.write_facts([fact])
        # Retrieve the generated icalendar.Event.
        result = ical_writer.calendar.add_component.call_args[0][0]
        assert result.get("dtstart").dt == fact.start
        assert result.get("dtend").dt == fact.end + datetime.timedelta(seconds=1)
        assert result.get("summary") == fact.activity_name
        # Make lists of [vText] and [str], else comparison fails.
        #  NO: assert result.get('categories') == fact.category
        assert list(result.get("categories").cats) == list([fact.category_name])
        assert result.get("categories").cats[0] == fact.category_name
        assert result.get("description") == fact.description_or_empty

    def test_ical_writer_write_report_not_implemented(self, ical_writer):
        with pytest.raises(NotImplementedError):
            ical_writer.write_report(table=[], headers=[])

    def test_ical_writer_write_facts_written(self, ical_writer, fact, path):
        """Make sure the calendar is actually written to disk before file is closed."""
        ical_writer.write_facts([fact])
        with open(path, "r", encoding="utf8") as fobj:
            # Create an icalendar.cal.Calendar from the file contents.
            result = Calendar.from_ical(fobj.read())
            assert result.walk()



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest


class TestCategoryManager:
    def test_save_new(self, basestore, category, mocker):
        """Make sure that saving an new category calls ``__add``."""
        mocker.patch.object(basestore.categories, "_add", return_value=category)
        try:
            basestore.categories.save(category)
        except NotImplementedError:
            pass
        assert basestore.categories._add.called

    def test_save_existing(self, basestore, category, mocker):
        category.pk = 0
        mocker.patch.object(basestore.categories, "_update", return_value=category)
        try:
            basestore.categories.save(category)
        except NotImplementedError:
            pass
        assert basestore.categories._update.called

    def test_save_wrong_type(self, basestore, category):
        with pytest.raises(TypeError):
            basestore.categories.save([])

    def test_get_or_create_existing(self, basestore, category, mocker):
        """Make sure the category is beeing looked up and no new one is created."""
        mocker.patch.object(basestore.categories, "get_by_name", return_value=category)
        mocker.patch.object(basestore.categories, "_add", return_value=category)
        try:
            basestore.categories.get_or_create(category.name)
        except NotImplementedError:
            pass
        assert basestore.categories._add.called is False
        assert basestore.categories.get_by_name.called

    def test_get_or_create_new_category(self, basestore, category, mocker):
        """Make sure the category is beeing looked up and new one is created."""
        mocker.patch.object(basestore.categories, "_add", return_value=category)
        mocker.patch.object(basestore.categories, "get_by_name", side_effect=KeyError)
        try:
            basestore.categories.get_or_create(category.name)
        except NotImplementedError:
            pass
        assert basestore.categories.get_by_name.called
        assert basestore.categories._add.called

    def test_add_not_implemented(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories._add(category)

    def test_update_not_implemented(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories._update(category)

    def test_remove_not_implemented(self, basestore, category):
        with pytest.raises(NotImplementedError):
            basestore.categories.remove(category)

    def test_get_not_implemented_invalid_pk(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get(12)

    def test_get_not_implemented_invalid_pk_type(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get_by_name("fooo")

    def test_get_all_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get_all()

    def test_get_all_by_usage_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.get_all_by_usage()

    def test_gather_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.categories.gather(query_terms=None)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest

from nark.backends.sqlalchemy.objects import AlchemyCategory
from nark.backends.sqlalchemy.storage import SQLAlchemyStore
from nark.config import decorate_config

# isort: off / Otherwise isort removes one of the following blank lines,
# but it doesn't if you remove the comment atop the following class.


class TestStore(object):
    """Tests to make sure our store/test setup behaves as expected."""

    # The reason we see a great deal of count == 0 statements is to make sure that
    # db rollback works as expected. Once we are confident in our sqlalchemy/pytest
    # setup those are not really needed.

    def test_build_is_not_persistent(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that calling ``factory.build()`` does not create a
        persistent db entry.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory.build()
        assert alchemy_store.session.query(AlchemyCategory).count() == 0

    def test_factory_call_persistent(self, alchemy_store, alchemy_category_factory):
        """Make sure that ``factory()`` does creates a persistent db entry."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_create_is_persistent(self, alchemy_store, alchemy_category_factory):
        """Make sure that  ``create()`` does creates a persistent db entry."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_build_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that factory instances have no pk assigned."""
        instance = alchemy_category_factory.build()
        assert instance.pk

    def test_create_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that factory.create instances have pk assigned."""
        instance = alchemy_category_factory.create()
        assert instance.pk

    def test_instance_fixture(self, alchemy_store, alchemy_category):
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert alchemy_category.pk
        assert alchemy_category.name

    def test_get_db_url(self, alchemy_config_parametrized, alchemy_store):
        """Make sure that db_url composition works as expected."""
        config, expectation = alchemy_config_parametrized
        alchemy_store.config = decorate_config(config)
        assert alchemy_store.db_url == expectation

    def test_get_db_url_missing_keys(
        self,
        alchemy_config_missing_store_config_parametrized,
        alchemy_store,
    ):
        """
        Make sure that db_url composition throws error if key/values are
        missing in config.
        """
        with pytest.raises(ValueError):
            alchemy_store.config = decorate_config(
                alchemy_config_missing_store_config_parametrized
            )
            # If decorate_config() does not raise on engine missing error,
            # db_url will raise on missing path, host, etc.
            alchemy_store.db_url

    def test_init_with_unicode_path(self, alchemy_config, db_path_parametrized):
        """Test that Instantiating a store with a unicode path works."""
        alchemy_config["db.path"] = db_path_parametrized
        assert SQLAlchemyStore(alchemy_config)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import re

from nark import get_version


class TestNarkGetVer:
    def test_get_version_argless(self):
        # E.g., '3.2.4.dev5+gd15b68dc.d20201209'
        pkg_version = get_version()
        assert re.match(r"^[0-9]+\.[0-9]+", pkg_version)

    def test_get_version_include_head_normal(self):
        # E.g., '3.2.4.dev5+gd15b68dc.d20201209 (3.2.4.dev16+g2bd6b40e)'
        pkg_version = get_version(include_head=True)
        assert re.match(r"^[0-9]+\.[0-9]+.* (.*)$", pkg_version)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Basic item class object generators."""

from pytest_factoryboy import register

from nark.tests import factories

# NOTE: To make the factories available to a test, import this file
#       from the test module or from a conftest along its path.
#       - You'll want to glob everything to get fact_factory, etc. E.g.,
#         Use this style import:
#           # Register the fact_factory, etc.
#           from nark.tests.item_factories import *  # noqa: F401, F403

# For generating backend store items, see the Alchemy item generators:
#
#   nark/tests/backends/sqlalchemy/conftest.py

register(factories.CategoryFactory)
register(factories.ActivityFactory)
register(factories.TagFactory)
register(factories.FactFactory)

# Note that registering the FactFactory, for example, is essentially:
#
#   @pytest.fixture
#   def fact_factory():
#       """Return a factory class that generates non-persisting Fact instances."""
#       return factories.FactFactory.build



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from ....managers.activity import BaseActivityManager
from ..objects import AlchemyActivity, AlchemyCategory, AlchemyFact
from . import query_apply_true_or_not
from .manager_base import BaseAlchemyManager

__all__ = ("ActivityManager",)


class ActivityManager(BaseAlchemyManager, BaseActivityManager):
    """ """

    def __init__(self, *args, **kwargs):
        super(ActivityManager, self).__init__(*args, **kwargs)

    # ***

    def get_or_create(self, activity, raw=False, skip_commit=False):
        """
        Custom version of the default method in order to provide access to
        Alchemy instances.

        Args:
            activity (nark.Activity): Activity we want.
            raw (bool): Whether to return the AlchemyActivity instead.

        Returns:
            nark.Activity: Activity.
        """

        self.store.logger.debug("Received: {!r} / raw: {}".format(activity, raw))

        try:
            result = self.get_by_composite(activity.name, activity.category, raw=raw)
        except KeyError:
            result = self._add(activity, raw=raw, skip_commit=skip_commit)
        self.store.logger.debug("Returning: {!r}".format(result))
        return result

    # ***

    def _add(self, activity, raw=False, skip_commit=False):
        """
        Add a new ``Activity`` instance to the databasse.

        Args:
            activity (nark.Activity): nark activity

        Returns:
            nark.Activity: nark activity representation of stored instance.

        Raises:
            ValueError: If the passed activity has a PK.
            ValueError: If the category/activity.name combination to be added is
                already present in the db.
        """
        self.adding_item_must_not_have_pk(activity)

        try:
            self.get_by_composite(activity.name, activity.category)
            message = _(
                "The database already contains that name/category.name combination."
            )
            self.store.logger.error(message)
            raise ValueError(message)
        except KeyError:
            pass

        alchemy_activity = AlchemyActivity(
            pk=None,
            name=activity.name,
            category=None,
            deleted=bool(activity.deleted),
            # FIXME/2020-05-19: Remove hidden...
            hidden=bool(activity.hidden),
        )
        if activity.category:
            try:
                category = self.store.categories.get_by_name(
                    activity.category.name, raw=True
                )
            except KeyError:
                category = AlchemyCategory(
                    pk=None,
                    name=activity.category.name,
                    deleted=bool(activity.category.deleted),
                    hidden=bool(activity.category.hidden),
                )
        else:
            category = None
        alchemy_activity.category = category

        result = self.add_and_commit(
            alchemy_activity,
            raw=raw,
            skip_commit=skip_commit,
        )

        return result

    # ***

    def _update(self, activity):
        """
        Update a given Activity.

        Args:
            activity (nark.Activity): Activity to be updated.

        Returns:
            nark.Activity: Updated activity.

        Raises:
            ValueError: If the new name/category.name combination is already taken.
            ValueError: If the the passed activity does not have a PK assigned.
            KeyError: If the the passed activity.pk can not be found.
        """

        self.store.logger.debug("Received: {!r}".format(activity))

        if not activity.pk:
            message = _(
                "The Activity passed ('{!r}') does not have a PK."
                " We do not know which entry to modify."
            ).format(activity)
            self.store.logger.error(message)
            raise ValueError(message)

        try:
            self.get_by_composite(activity.name, activity.category)
            message = _(
                "The database already contains that Activity and Category combination."
            )
            self.store.logger.error(message)
            raise ValueError(message)
        except KeyError:
            # I.e., 'No activity named ‘<>’ of category ‘<>’ was found.'
            pass

        alchemy_activity = self.store.session.query(AlchemyActivity).get(activity.pk)
        if not alchemy_activity:
            message = _("No Activity with PK ‘{}’ was found.").format(activity.pk)
            self.store.logger.error(message)
            raise KeyError(message)
        alchemy_activity.name = activity.name
        alchemy_activity.category = self.store.categories.get_or_create(
            activity.category,
            raw=True,
        )
        alchemy_activity.deleted = bool(activity.deleted)
        try:
            self.store.session.commit()
        except IntegrityError as err:
            # (lb): I think this path unreachable, because get_by_composite should
            # find it first. Or is it something else?
            message = _(
                "There's already an activity like this for the given category."
                " Cannot change this activity's values. Original exception: {}"
            ).format(str(err))
            self.store.logger.error(message)
            raise ValueError(message)
        result = alchemy_activity.as_hamster(self.store)
        self.store.logger.debug("Returning: {!r}".format(result))
        return result

    # ***

    def remove(self, activity):
        """
        Remove an activity from our internal backend.

        Args:
            activity (nark.Activity): The activity to be removed.

        Returns:
            bool: True

        Raises:
            KeyError: If the given ``Activity`` can not be found in the database.
        """

        self.store.logger.debug("Received: {!r}".format(activity))

        if not activity.pk:
            message = _(
                "Activity ‘{!r}’ has no PK. Are you trying to remove a new Activity?"
            ).format(activity)
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_activity = self.store.session.query(AlchemyActivity).get(activity.pk)
        if not alchemy_activity:
            message = _("No Activity with PK ‘{}’ was found.").format(activity.pk)
            self.store.logger.error(message)
            raise KeyError(message)

        if alchemy_activity.facts:
            # FIXME: Untested. (lb): Also, this is how deleted is meant to be used...
            #        Not quite sure I see how it'd work... better workflow should
            #        involve reassigning all Facts that had one Activity to another,
            #        otherwise you end up with weird orphan Activities that the UX
            #        doesn't really process well.
            alchemy_activity.deleted = True
            self.store.activities._update(alchemy_activity)
        else:
            self.store.session.delete(alchemy_activity)
        self.store.session.commit()
        self.store.logger.debug("Deleted: {!r}".format(activity))

    # ***

    def get(self, pk, deleted=None, raw=False):
        """
        Query for an Activity with given key.

        Args:
            pk: PK to look up.
            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Activity: Activity with given PK.

        Raises:
            KeyError: If no such pk was found.
        """

        self.store.logger.debug("Received PK: ‘{}’ / raw: {}.".format(pk, raw))

        if deleted is None:
            result = self.store.session.query(AlchemyActivity).get(pk)
        else:
            query = self.store.session.query(AlchemyActivity)
            query = query.filter(AlchemyActivity.pk == pk)
            query = query_apply_true_or_not(query, AlchemyActivity.deleted, deleted)
            results = query.all()
            assert len(results) <= 1
            result = results[0] if results else None

        if not result:
            message = _("No Activity with PK ‘{}’ was found.").format(pk)
            self.store.logger.error(message)
            raise KeyError(message)
        if not raw:
            result = result.as_hamster(self.store)
        self.store.logger.debug("Returning: {!r}.".format(result))
        return result

    # ***

    # NOTE: Unlike Category and Tag, there is no Activity.get_by_name.

    # ***

    def get_by_composite(self, name, category, raw=False):
        """
        Retrieve an activity by its name and category.

        Args:
            name (str): The activities name.
            category (nark.Category or None): The activities category.
                May be None.
            raw (bool): Return the AlchemyActivity instead.

        Returns:
            nark.Activity: The activity if it exists in this combination.

        Raises:
            KeyError: if composite key can not be found in the db.

        Note:
            As far as we understand the legacy code in ``__change_category``
            and ``__get_activity_by`` the combination of activity.name and
            activity.category is unique. This is reflected in the uniqueness
            constraint of the underlying table.
        """

        self.store.logger.debug(
            "Received: {!r} / name: ‘{}’ / raw: {}".format(category, name, raw)
        )

        if category:
            category = category.name
            try:
                alchemy_category = self.store.categories.get_by_name(category, raw=True)
            except KeyError:
                message = _(
                    "The category passed ({}) does not exist in the backend. "
                    "Consequently no related activity can be returned."
                ).format(category)
                # (lb): This was error, but shouldn't be; callers catch if they care.
                self.store.logger.debug(message)
                raise KeyError(message)
        else:
            alchemy_category = None

        # EXPLAIN: (lb): Is name ever not a string here?
        name = str(name)
        try:
            query = self.store.session.query(AlchemyActivity)
            # Note that if alchemy_category is None -- because caller passed None --
            # then this only finds an Activity if it has no Category.
            query = query.filter_by(name=name).filter_by(category=alchemy_category)
            result = query.one()
        except NoResultFound:
            message = _(
                "No Activity named ‘{name}’ of category ‘{category}’ was found."
            ).format(name=name, category=category)
            self.store.logger.debug(message)
            raise KeyError(message)
        if not raw:
            result = result.as_hamster(self.store)
        self.store.logger.debug("Returning: {!r}.".format(result))
        return result

    # ***
    # *** gather() call-outs (used by get_all/get_all_by_usage).
    # ***

    @property
    def _gather_query_alchemy_cls(self):
        return AlchemyActivity

    @property
    def _gather_query_order_by_name_col(self):
        return "activity"

    def _gather_query_start_timeless(self, qt, alchemy_cls):
        # Query on AlchemyActivity.
        query = super(ActivityManager, self)._gather_query_start_timeless(
            qt, alchemy_cls
        )

        query = self._gather_query_join_category(qt, query)
        return query

    def _gather_query_start_aggregate(self, qt, agg_cols):
        query = self.store.session.query(AlchemyFact, *agg_cols)
        query = query.outerjoin(AlchemyFact.activity)
        query = self._gather_query_join_category(qt, query)
        return query

    def _gather_query_join_category(self, qt, query):
        # We'll usually join the Category table now, so that we can get all
        # the data we need in one query, to avoid SQLAlchemy lazy-loading
        # the Category separately for each Activity.
        # - If we don't join Category now, SQLAlchemy will fetch the Category
        #   the first time we reference activity.category.
        # - It not raw, query_process_results will call as_hamster on the
        #   Activity, which will reference activity.category. So join now
        #   if not qt.raw and avoid fetching the Category separately for
        #   each Activity.
        # - If sorting on Category, we'll need to join the table.
        # - If matching on Category, we'll need to join the table.
        requires_category_table = (
            not qt.raw or qt.match_categories or qt.sort_cols_has_any("category")
        )
        if requires_category_table:
            query = query.outerjoin(AlchemyCategory)

        return query

    # ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Provides string-related functional methods."""

from ansiwrap_hotoffthehamster import shorten

__all__ = ("format_value_truncate",)


# (lb): Note that textwrap/ansiwrap.shorten only truncates on whitespace or hyphens,
# so the actual truncated (shortened) text will not be exactly trunc_width, but will
# instead be *at most* trunc_width.
# - We could alternatively chop exactly at trunc_width, e.g.,
#     val = val[:trunc_width - len(ellipsis)] + ellipsis
#   which would truncate inside words, e.g., "it would trunca...".
#   But this approach only works on ANSI-free text, and we want this function to work
#   with ANSI-encoded (e.g., colorful) output text. So we use the shorten function.
# Note, too, that regardless of trunc_width, this function *always* replaces each
# newline with its escape sequence ("\n"), because it assumes, if it's being called,
# that the caller wants to at least condense the output (fit it all on one line).
def format_value_truncate(val, trunc_width=None):
    if not val:
        return val
    val = "\\n".join(str(val).splitlines())
    if trunc_width is not None and trunc_width > 0:
        # textwrap3 raises ValueError if max shorter than placeholder.
        ellipsis = "..."
        trunc_width = max(trunc_width, len(ellipsis))
        val = shorten(val, trunc_width, placeholder=ellipsis)
    return val



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures that are of general use."""

import datetime

import pytest

from nark.tests import factories

# Register the category_factory, etc.
from nark.tests.item_factories import *  # noqa: F401, F403

# ***

# Fixtures for: tests/items/test_activity.py.


# Categories
@pytest.fixture(
    params=(
        None,
        True,
    )
)
def category_valid_parametrized(
    request,
    category_factory,
    name_string_valid_parametrized,
):
    """Provide a variety of valid category fixtures."""
    if request.param:
        result = category_factory(name=name_string_valid_parametrized)
    else:
        result = None
    return result


# +++

# (lb): Unused.


@pytest.fixture
def category_valid_parametrized_without_none(
    request,
    category_factory,
    name_string_valid_parametrized,
):
    """
    Provide a parametrized category fixture but not ``None``.

    This fixuture will represent a wide array of potential name charsets as well
    but not ``category=None``.
    """
    return category_factory(name=name_string_valid_parametrized)


# Activities
@pytest.fixture
def activity_valid_parametrized(
    request,
    activity_factory,
    name_string_valid_parametrized,
    category_valid_parametrized,
    deleted_valid_parametrized,
):
    """Provide a huge array of possible activity versions. Including None."""
    return activity_factory(
        name=name_string_valid_parametrized,
        category=category_valid_parametrized,
        deleted=deleted_valid_parametrized,
    )


@pytest.fixture
def new_activity_values(category):
    """Return garanteed modified values for a given activity."""

    def modify(activity):
        return {
            "name": activity.name + "foobar",
        }

    return modify


# ***

# Fixtures for: tests/items/test_fact.py.


@pytest.fixture
def fact():
    """Provide a randomized non-persistant Fact-instance."""
    return factories.FactFactory.build()


# +++


@pytest.fixture(params=("%M", "%H:%M", "HHhMMm", ""))
def string_delta_style_parametrized(request):
    """Provide all possible format option for ``Fact().format_delta()``."""
    return request.param


# ***

# (lb): Unused.


@pytest.fixture
def today_fact(fact_factory):
    """Return a ``Fact`` instance that start and ends 'today'."""
    # MAYBE: Use controller.store.now ?
    start = datetime.datetime.utcnow()
    end = start + datetime.timedelta(minutes=30)
    return fact_factory(start=start, end=end)


@pytest.fixture
def not_today_fact(fact_factory):
    """Return a ``Fact`` instance that neither start nor ends 'today'."""
    # MAYBE: Use controller.store.now ?
    start = datetime.datetime.utcnow() - datetime.timedelta(days=2)
    end = start + datetime.timedelta(minutes=30)
    return fact_factory(start=start, end=end)


@pytest.fixture
def current_fact(fact_factory):
    """Provide a ``ongoing fact``, which has a start time but no end time."""
    # MAYBE: Use controller.store.now ?
    return fact_factory(
        start=datetime.datetime.utcnow().replace(microsecond=0),
        end=None,
    )



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from freezegun import freeze_time

from nark.helpers import fact_time
from nark.helpers.parse_errors import ParserInvalidDatetimeException
from nark.helpers.parse_time import HamsterTimeSpec, parse_dated, parse_datetime_human


class TestGetDayEnd(object):
    @pytest.mark.parametrize(
        ("day_start", "expectation"),
        [
            (datetime.time(0, 0, 0), datetime.time(23, 59, 59)),
            (datetime.time(5, 30, 0), datetime.time(5, 29, 59)),
            (datetime.time(23, 59, 59), datetime.time(23, 59, 58)),
            (datetime.time(14, 44, 23), datetime.time(14, 44, 22)),
        ],
    )
    def test_various_day_start_times(self, base_config, day_start, expectation):
        """Ensure that resulting end times match our expectation given ``day_start``-"""
        base_config["time.day_start"] = day_start
        assert fact_time.day_end_time(day_start) == expectation


class TestEndDayToDatetime(object):
    @pytest.mark.parametrize(
        ("day_start", "expectation"),
        [
            (datetime.time(0, 0, 0), datetime.datetime(2015, 4, 15, 23, 59, 59)),
            (datetime.time(5, 30, 0), datetime.datetime(2015, 4, 16, 5, 29, 59)),
            (datetime.time(23, 59, 59), datetime.datetime(2015, 4, 16, 23, 59, 58)),
            (datetime.time(14, 44, 23), datetime.datetime(2015, 4, 16, 14, 44, 22)),
        ],
    )
    def test_same_end_day_various_day_starts(self, day_start, expectation):
        """
        Ensure that resulting ``end datetimes`` match our expectation given ``day_end``
        """
        end_day = datetime.datetime(2015, 4, 15)
        assert fact_time.day_end_datetime(end_day, day_start) == expectation


# ***


class TestParseTimeHamsterTimeSpec(object):
    """Tests the parse_time module's HamsterTimeSpec class."""

    # Note that much of HamsterTimeSpec indirectly tested via test_parsing.py.

    def test_cannot_instantiate(self):
        with pytest.raises(NotImplementedError):
            HamsterTimeSpec()

    def test_has_time_of_day_false(self):
        has_tod = HamsterTimeSpec.has_time_of_day("2015-12-12")
        assert not has_tod

    def test_has_time_of_day_true(self):
        has_tod = HamsterTimeSpec.has_time_of_day("2015-12-12 18:55")
        assert has_tod


@freeze_time("2015-12-10 12:30")
class TestParseTimeFunctions(object):
    """Tests the parse_time module."""

    # Note that much of parse_time.py indirectly tested via test_parsing.py.

    @pytest.mark.parametrize(
        ("time", "expectation"),
        [
            ("18:55", datetime.datetime(2015, 12, 9, 18, 55)),
            ("18:55:34", datetime.datetime(2015, 12, 9, 18, 55, 34)),
            ("2014-12-10", datetime.datetime(2014, 12, 10, 0, 0)),
            ("2015-10-02 18:12", datetime.datetime(2015, 10, 2, 18, 12)),
            ("2015-10-02 18:12:33", datetime.datetime(2015, 10, 2, 18, 12, 33)),
            # Relative time.
            ("-60", datetime.datetime(2015, 12, 10, 11, 30)),
            ("+2h", datetime.datetime(2015, 12, 10, 14, 30)),
            # Friendly time.
            ("Monday", datetime.datetime(2015, 12, 7, 0, 0)),
        ],
    )
    def test_parse_dated_valid_times(self, time, expectation):
        """Make sure that given times are parsed as expected."""
        parsed = parse_dated(time, time_now=datetime.datetime.now())
        assert parsed == expectation

    @pytest.mark.parametrize(
        "time",
        [
            "18 55",
            "18:555",
            "18:66",
            "2014 01 04 12:30",
            "2014-13-13 12:30",
        ],
    )
    def test_parse_dated_invalid_times(self, time):
        """Ensure that invalid times throw an exception."""
        with pytest.raises(ParserInvalidDatetimeException):
            # F841 local variable 'parsed_' is assigned to but never used
            parsed_ = parse_dated(  # noqa: F841
                time,
                time_now=datetime.datetime.now(),
                cruftless=True,
            )
            assert False  # Unreachable.

    def test_parse_dated_not_a_str(self):
        not_a_str = object()
        parsed = parse_dated(not_a_str, time_now=datetime.datetime.now())
        assert parsed is not_a_str

    def test_parse_datetime_human_local_tz_True(self):
        # FIXME: (lb): There's a timezone feature coming. I swear. One of these years.
        #        For now, can at least test those 2 lines that production does not use.
        # Acceptable time zones, e.g.: '+0500', 'US/Eastern', datetime.timezone.utc.
        # You can also specify tzinfo in datetime.
        parsed = parse_datetime_human(
            "Monday",
            time_now=datetime.datetime.now(),
            local_tz="UTC",
        )
        expected = datetime.datetime(2015, 12, 7, 0, 0, tzinfo=datetime.timezone.utc)
        assert parsed == expected


# ***


class TestValidateStartEndRange(object):
    """Unittests for validation function."""

    @pytest.mark.parametrize(
        "range",
        (
            (
                datetime.datetime(2016, 12, 1, 12, 30),
                datetime.datetime(2016, 12, 1, 12, 45),
            ),
            (
                datetime.datetime(2016, 1, 1, 12, 30),
                datetime.datetime(2016, 12, 1, 12, 45),
            ),
            (
                datetime.datetime(2016, 1, 1, 12, 30),
                datetime.datetime(2016, 12, 1, 1, 45),
            ),
        ),
    )
    def test_valid_ranges(self, range):
        """Make sure that ranges with end > start pass validation."""
        result = fact_time.must_not_start_after_end(range)
        assert result == range

    @pytest.mark.parametrize(
        "range",
        (
            (
                datetime.datetime(2016, 12, 1, 12, 30),
                datetime.datetime(2016, 12, 1, 10, 45),
            ),
            (
                datetime.datetime(2016, 1, 13, 12, 30),
                datetime.datetime(2016, 1, 1, 12, 45),
            ),
        ),
    )
    def test_invalid_ranges(self, range):
        """Make sure that ranges with start > end fail validation."""
        with pytest.raises(ValueError):
            fact_time.must_not_start_after_end(range)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""ICAL writer output format module."""

import datetime

import lazy_import

from . import ReportWriter

__all__ = ("ICALWriter",)

# Profiling: load icalendar: ~ 0.008 secs.
icalendar = lazy_import.lazy_module("icalendar")


class ICALWriter(ReportWriter):
    """A simple ical writer for fact export."""

    def __init__(self, *args, **kwargs):
        """
        Initiate new instance and open an output file like object.

        Args:
            path: File like object to be opend. This is where all output
                will be directed to. datetime_format (str): String specifying
                how datetime information is to be rendered in the output.
        """
        kwargs["output_b"] = True
        super(ICALWriter, self).__init__(*args, **kwargs)
        self.calendar = icalendar.Calendar()

    def _write_fact(self, idx, fact):
        """
        Write a singular fact to our report.

        Note:
            * ``dtend`` is non-inclusive according to Page 54 of RFC 5545

        Returns:
            None: If everything worked out alright.
        """
        # [FIXME]
        # It apears that date/time requirements for VEVENT have changed between
        # RFCs. 5545 now seems to require a 'dstamp' and a 'uid'!

        # (lb): I'm not sure the utility of this export format without
        # ever personally having tested importing it anywhere. For
        # instance, maybe exporting the Activity@Category is a better
        # idea than using the separate "categories" and "summary" fields
        # (and I'm not sure how those fields are intended to be used, nor
        # how different applications might show their values). In any case,
        # I've maintained this format because it existed in hamster-lib, and
        # it seems like something people might want.

        event = icalendar.Event()

        event.add("dtstart", fact.start)
        if fact.end:
            # MAGIC_NUMBER: (lb): Add one second, because `dtend` is non-inclusive.
            event.add("dtend", fact.end + datetime.timedelta(seconds=1))
        event.add("categories", [fact.category_name])
        event.add("summary", fact.activity_name)
        event.add("description", fact.description_or_empty)

        self.calendar.add_component(event)

    def write_report(self, table, headers, tabulation=None):
        raise NotImplementedError

    def _close(self):
        """Custom close method to make sure the calendar is actually writen do disk."""
        self.output_file.write(self.calendar.to_ical())
        return super(ICALWriter, self)._close()



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""
Base classes for implementing storage backends.

Note:
    * This is propably going to be replaced by a ``ABC``-bases solution.
    * Basic validation could be done here then. This would mean we just need
      to test them once and our actual backends focus on the CRUD implementation.
"""

import datetime
from gettext import gettext as _

from ..helpers.parse_time import parse_dated
from ..items.item_base import BaseItem
from .query_terms import QueryTerms

__all__ = ("BaseManager",)


class BaseManager(object):
    """Base class for all object managers."""

    def __init__(self, store):
        self.store = store

    @property
    def config(self):
        return self.store.config

    # ***

    def adding_item_must_not_have_pk(self, hamster_item):
        message = _("Adding item: {!r}.".format(hamster_item))
        self.store.logger.debug(message)
        if not hamster_item.pk:
            return
        message = _(
            "The {} item ({!r}) cannot be added because it already has a PK."
            " Perhaps call the ``_update`` method instead".format(
                self.__class__.__name__,
                hamster_item,
            )
        )
        self.store.logger.error(message)
        raise ValueError(message)

    # ***

    def save(self, item, cls=BaseItem, named=False, **kwargs):
        """
        Save a Nark object instance to user's selected backend.

        Will either ``_add`` or ``_update`` based on item PK.

        Args:
            tag (nark.BaseItem, i.e., Activity/Category/Fact/Tag):
                Nark instance to be saved.

        Returns:
            nark.BaseItem: Saved Nark instance.

        Raises:
            TypeError: If the ``item`` parameter is not a valid ``BaseItem`` instance.
        """

        if not isinstance(item, cls):
            message = _("You need to pass a {} object").format(cls.__name__)
            self.store.logger.debug(message)
            raise TypeError(message)

        # (lb): Not sure this is quite what we want, but Activity has been doing this,
        # and I just made this base class, so now all items will be doing this.
        if named and not item.name:
            raise ValueError(_("You must specify an item name."))

        self.store.logger.debug(_("'{}' has been received.".format(item)))

        # NOTE: Not assuming that PK is an int, i.e., not testing '> 0'.
        #       (Also, if pk really 0, this raises ValueError.)
        if item.pk or item.pk == 0:
            result = self._update(item, **kwargs)
        else:
            # PK is empty string, empty list, None, etc., but not 0.
            result = self._add(item, **kwargs)
        return result

    # ***

    def _gather_prepare_query_terms(self, query_terms, **kwargs):
        if query_terms is not None:
            return query_terms, kwargs
        return QueryTerms(**kwargs), {}

    # ***

    def get_all(self, query_terms=None, **kwargs):
        """
        Return all items and any requested stats matching the given search criteria.

        Returns:
            list: List of results matching given specifications. Each results includes
            an item or aggregated item and additional calculations as specified by the
            query terms.

        Raises:
            TypeError: If ``since`` or ``until`` are not ``datetime.date``,
                ``datetime.time`` or ``datetime.datetime`` objects.

            ValueError: If ``until`` is before ``since``.
        """
        qt, kwargs = self._gather_prepare_query_terms(query_terms, **kwargs)

        def _get_all():
            # MAYBE: get_all has a side-effect of mutating qt.since and qt.until,
            #   which is not necessarily desirable.
            qt.since, qt.until = _must_parse_since_until(qt.since, qt.until)
            return self.gather(qt, **kwargs)

        def _must_parse_since_until(since, until):
            self.store.logger.debug("since: {} / until: {}".format(since, until))

            # Convert the since and until time strings to datetimes.
            since = parse_dated(since, self.store.now) if since else None
            until = parse_dated(until, self.store.now) if until else None

            since_dt = _must_verify_since(since)
            until_dt = _must_verify_until(until)
            if since_dt and until_dt and (until_dt <= since_dt):
                message = _("`until` cannot be earlier than `since`.")
                self.store.logger.debug(message)
                raise ValueError(message)

            return since_dt, until_dt

        def _must_verify_since(since):
            if since is None:
                return since

            if isinstance(since, datetime.datetime):
                # isinstance(datetime.datetime, datetime.date) returns True,
                # which is why we need to catch this case first.
                since_dt = since
            elif isinstance(since, datetime.date):
                # The user specified a date, but not a time. Assume midnight.
                # MAYBE: Use config['day_start'] and subtract a day minus a minute?
                self.store.logger.debug(
                    _("Using midnight as clock time for `since` date.")
                )
                day_start = self.config["time.day_start"]
                since_dt = datetime.datetime.combine(since, day_start)
            elif isinstance(since, datetime.time):
                since_dt = datetime.datetime.combine(datetime.date.today(), since)
            else:
                message = _(
                    "Unable to convert the since input to a datetime."
                    " Neither date, nor time, nor datetime: ‘{}’".format(str(since))
                )
                self.store.logger.debug(message)
                raise TypeError(message)
            return since_dt

        def _must_verify_until(until):
            if until is None:
                return until

            if isinstance(until, datetime.datetime):
                # isinstance(datetime.datetime, datetime.date) returns True,
                # which is why we need to except this case first.
                until_dt = until
            elif isinstance(until, datetime.date):
                # MAYBE: (lb): Feels weird that since defaults to midnight,
                #   but until defaults to 'day_start' plus a day...
                #   (need to TESTME to really feel what's going on).
                until_dt = self.day_end_datetime(until)
            elif isinstance(until, datetime.time):
                until_dt = datetime.datetime.combine(datetime.date.today(), until)
            else:
                message = _(
                    "Unable to convert the until input to a datetime."
                    " Neither date, nor time, nor datetime: ‘{}’".format(str(until))
                )
                raise TypeError(message)
            return until_dt

        return _get_all()

    # ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import copy

import pytest

from nark.items.category import Category


class TestCategory(object):
    @classmethod
    def as_repr(cls, category):
        if category is None:
            return repr(category)
        repr_f = "Category(_name={name}, deleted={deleted}, hidden={hidden}, pk={pk})"
        repred = repr_f.format(
            pk=repr(category.pk),
            name=repr(category.name),
            deleted=repr(category.deleted),
            hidden=repr(category.hidden),
        )
        return repred

    def test_init_valid(self, name_string_valid_parametrized, pk_valid_parametrized):
        """Make sure that Category constructor accepts all valid values."""
        category = Category(name_string_valid_parametrized, pk_valid_parametrized)
        assert category.name == name_string_valid_parametrized
        assert category.pk == pk_valid_parametrized

    def test_init_invalid(self, name_string_invalid_parametrized):
        """Make sure that Category constructor rejects all invalid values."""
        with pytest.raises(ValueError):
            Category(name_string_invalid_parametrized)

    def test_as_tuple_include_pk(self, category):
        """
        Make sure categories tuple representation works as intended and pk
        is included.
        """
        deleted = False
        hidden = False
        assert category.as_tuple() == (category.pk, category.name, deleted, hidden)

    def test_as_tuple_exclude_pf(self, category):
        """
        Make sure categories tuple representation works as intended and pk
        is excluded.
        """
        deleted = False
        hidden = False
        cat_tuple = category.as_tuple(include_pk=False)
        our_tuple = (False, category.name, deleted, hidden)
        assert cat_tuple == our_tuple

    def test_equal_fields_true(self, category):
        """
        Make sure that two categories that differ only in their PK compare equal.
        """
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        assert category.equal_fields(other_category)

    def test_equal_fields_false(self, category):
        """
        Make sure that two categories that differ not only in their PK
        compare unequal.
        """
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        other_category.name += "foobar"
        assert category.equal_fields(other_category) is False

    def test__eq__false(self, category):
        """Make sure that two distinct categories return ``False``."""
        other_category = copy.deepcopy(category)
        other_category.pk = 1
        assert category is not other_category
        assert category != other_category

    def test__eq__true(self, category):
        """Make sure that two identical categories return ``True``."""
        other_category = copy.deepcopy(category)
        assert category is not other_category
        assert category == other_category

    def test_is_hashable(self, category):
        """Test that ``Category`` instances are hashable."""
        assert hash(category)

    def test_hash_method(self, category):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(category) == hash(category.as_tuple())

    def test_hash_different_between_instances(self, category_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Category.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(category_factory()) != hash(category_factory())

    def test__str__(self, category):
        """Test string representation."""
        assert "{name}".format(name=category.name) == str(category)

    def test__repr__(self, category):
        """Test representation method."""
        result = repr(category)
        assert isinstance(result, str)
        expectation = TestCategory.as_repr(category)
        assert result == expectation



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures that are of general use."""

import pytest
from configobj import ConfigObj

from nark.config.log_levels import LOG_LEVELS


@pytest.fixture(
    params=list(LOG_LEVELS.keys())
    + [
        123,
    ]
)
def log_level_valid_parametrized(request):
    """Return each of the valid log level strings."""
    return request.param


@pytest.fixture(params=(None, "123", "abc", ""))
def log_level_invalid_parametrized(request):
    """Return selection of invalid log level strings."""
    return request.param


@pytest.fixture
def configobj_instance(request):
    """Provide a ``ConfigObj`` instance and its expected config dict."""

    config = ConfigObj()
    config["db"] = {}
    config["db"]["orm"] = "sqlalchemy"
    config["db"]["engine"] = "sqlite"
    config["db"]["path"] = "/tmp/nark-tests-config.db"
    config["db"]["host"] = "www.example.com"
    config["db"]["port"] = 22
    config["db"]["name"] = "hamster"
    config["db"]["user"] = "hamster"
    config["db"]["password"] = "hamster"
    config["dev"] = {}
    config["dev"]["catch_errors"] = False
    config["dev"]["lib_log_level"] = "WARNING"
    config["dev"]["sql_log_level"] = "debug"
    config["time"] = {}
    config["time"]["allow_momentaneous"] = False
    config["time"]["day_start"] = "05:00:00"
    config["time"]["fact_min_delta"] = 60

    expectation = {
        "db": {
            "orm": "sqlalchemy",
            "engine": "sqlite",
            "path": "/tmp/nark-tests-config.db",
            "host": "www.example.com",
            "port": "22",
            "name": "hamster",
            "user": "hamster",
            "password": "hamster",
        },
        "dev": {
            # Devmode catch_errors could be deadly under test, as it sets a trace trap.
            "catch_errors": "False",
            "lib_log_level": "WARNING",
            "sql_log_level": "debug",
        },
        "time": {
            # allow_momentaneous is hidden, so won't show in as_dict, etc.
            #  'allow_momentaneous': 'False',
            "day_start": "05:00:00",
            "fact_min_delta": "60",
            # MAYBE: (lb): Consider fiddling with day_start and fact_min_delta
            # in specific tests and leaving them set to factory defaults here:
            #   'day_start': '',
            #   'fact_min_delta': 0,
        },
    }

    return config, expectation
